import abc
from functools import lru_cache

from .util import ttl_cache


class DBExtract():
    def __init__(self, datasets, circles=None, collections=None):
        """User-convenient access to dataset search results

        Parameters
        ----------
        datasets: list
            List of CKAN package dictionaries
        """
        self._circles = circles
        self._collections = collections
        self._dataset_name_index = None
        self.datasets = datasets

    @property
    @lru_cache(maxsize=1)
    def circles(self):
        if self._circles is None:
            cl = []
            for dd in self.datasets:
                name = dd["organization"]["name"]
                if name not in cl:
                    cl.append(name)
            self._circles = sorted(cl)
        return self._circles

    @property
    @lru_cache(maxsize=1)
    def collections(self):
        if self._collections is None:
            ct = []
            for dd in self.datasets:
                for gg in dd["groups"]:
                    name = gg["name"]
                    if gg not in ct:
                        ct.append(name)
            self._collections = sorted(ct)
        return self._collections

    def _generate_index(self):
        if self._dataset_name_index is None:
            index = {}
            for ii, dd in enumerate(self.datasets):
                index[dd["name"]] = ii
            self._dataset_name_index = index

    def get_dataset_dict(self, dataset_name):
        self._generate_index()
        return self.datasets[self._dataset_name_index[dataset_name]]


class DBModel():
    def __init__(self, interrogator, mode="public", *args, **kwargs):
        """User-convenient interface for DCOR/CKAN DBInterrogator

        Parameters
        ----------
        interrogator: DBInterrogator
            Class interacting with the database.
        mode: str
            This can be either "public" (only public datasets are
            considered) or "user" (datasets are limited to those owned
            by or shared with the user, including private datasets).
        """
        self.db = interrogator
        self.user_data = self.db.get_user_data()
        if mode not in ["public", "user"]:
            raise ValueError(
                "`mode` must be 'user', or 'public'; got '{}'!".format(mode))
        self.mode = mode
        self.search_query = {}
        self.set_filter()  # initiate search_query

    @property
    def is_up_to_date(self):
        return self.db.is_up_to_date()

    def get_circles(self):
        """Return the list of DCOR Circles"""
        return self.db.get_circles(mode=self.mode)

    def get_collections(self):
        """Return the list of DCOR Collections"""
        return self.db.get_collections(mode=self.mode)

    def get_user_datasets(self):
        """Return DBExtract with data owned by or shared with the user"""
        owned = self.db.get_datasets_user_owned()
        shared = self.db.get_datasets_user_shared()
        following = self.db.get_datasets_user_following()
        return DBExtract(owned+shared+following)

    def get_users(self):
        """Return the list of DCOR users"""
        return self.db.get_users()

    def search_dataset(self, query):
        """Search for a string in the database"""
        data = self.db.search_dataset(
            query,
            mode=self.mode,
            circles=self.search_query["circles"],
            collections=self.search_query["collections"],
        )
        extract = DBExtract(data,
                            circles=self.search_query["circles"],
                            collections=self.search_query["collections"],
                            )
        return extract

    def set_filter(self, circles=[], collections=[]):
        """Set a search query filter"""
        db_circles = self.db.get_circles(mode=self.mode)
        db_collections = self.db.get_collections(mode=self.mode)

        # If no circles or collections are specified, then
        # there would be no search results. For user
        # convenience, we set the complete lists instead.
        if not circles:
            circles = db_circles
        if not collections:
            collections = db_collections

        # Sanity checks
        for ci in circles:
            if ci not in db_circles:
                raise KeyError(
                    "Circle '{}' does not exist in '{}'!".format(ci, self.db))
        for co in collections:
            if co not in db_collections:
                raise KeyError(
                    "Collection '{}' does not exist in '{}'!".format(co,
                                                                     self.db))

        self.search_query["circles"] = circles
        self.search_query["collections"] = collections


class DBInterrogator(abc.ABC):
    def __init__(self, *args, **kwargs):
        self.search_query = {}

    @property
    def is_up_to_date(self):
        "Checks whether the local database copy is up-to-date"
        if self.local_version_score != self.remote_version_score:
            uptodate = False
        elif self.local_timestamp != self.remote_timestamp:
            uptodate = False
        else:
            uptodate = True
        return uptodate

    @abc.abstractmethod
    def get_circles(self, mode="public"):
        """Return the list of DCOR Circles"""
        pass

    @abc.abstractmethod
    def get_collections(self, mode="public"):
        """Return the list of DCOR Collections"""
        pass

    @abc.abstractmethod
    def get_datasets_user_following(self):
        """Return all datasets the user is following"""
        pass

    @abc.abstractmethod
    def get_datasets_user_owned(self):
        """Return all datasets owned by the user"""
        pass

    @abc.abstractmethod
    def get_datasets_user_shared(self):
        """Return all datasets shared with the user"""
        pass

    @abc.abstractmethod
    def get_user_data(self):
        """Return the current user data dictionary"""
        pass

    @abc.abstractmethod
    def get_users(self):
        """Return the list of DCOR users"""
        pass

    @abc.abstractmethod
    def search_dataset(self, circles, collections, mode="public"):
        pass

    @property
    @abc.abstractmethod
    def local_version_score(self):
        """Local database version"""

    @property
    @abc.abstractmethod
    def local_timestamp(self):
        """Local database date in seconds since epoch"""

    @property
    @ttl_cache(seconds=5)
    def remote_version_score(self):
        """Remote database version"""
        return 0

    @property
    @ttl_cache(seconds=5)
    def remote_timestamp(self):
        """Remote database date in seconds since epoch"""
        return 0



"""https://stackoverflow.com/questions/39304951/"""
from functools import wraps

from PyQt5 import QtCore


class Runner(QtCore.QThread):
    """Runs a function in the background"""

    def __init__(self, target, *args, **kwargs):
        super().__init__()
        self._target = target
        self._args = args
        self._kwargs = kwargs

    def run(self):
        self._target(*self._args, **self._kwargs)


def run_async(func):
    """Decorator for running a function in the background"""
    @wraps(func)
    def async_func(*args, **kwargs):
        runner = Runner(func, *args, **kwargs)
        # Keep the runner somewhere or it will be destroyed
        func.__runner = runner
        runner.start()

    return async_func



import copy
import json

import requests

from ._version import version


#: List of license lists for each DCOR server
SERVER_LICENCES = {}

#: List of supplementary resource schema dictionaries
SERVER_RSS = {}

#: List of supported resource suffixes
SERVER_RSUFFIX = {}


class APIError(BaseException):
    pass


class APIAuthorizationError(APIError):
    pass


class APIBadGatewayError(APIError):
    pass


class APIConflictError(APIError):
    pass


class APIGatewayTimeoutError(APIError):
    pass


class APIKeyError(APIError):
    pass


class APINotFoundError(APIError):
    pass


class CKANAPI:
    def __init__(self, server, api_key, ssl_verify=True):
        """User-convenient interface to the CKAN API"""
        self.api_key = api_key
        self.server = self._make_server_url(server)
        self.api_url = self._make_api_url(server)
        self.headers = {"X-CKAN-API-Key": api_key,
                        "user-agent": "DCOR-Aid/{}".format(version)
                        }
        self.verify = ssl_verify

    def _make_api_url(self, url):
        """Generate a complete CKAN API URL

        Any given string is changed to yield the
        form "https://domain.name/api/3/action".
        """
        url = self._make_server_url(url)
        if not url.endswith("/action/"):
            url = url.rstrip("/") + "/api/3/action/"
        return url

    def _make_server_url(self, url):
        """Generate a complete CKAN server URL

        Any given string is changed to yield the
        form "https://domain.name/".
        """
        if not url.count("//"):
            url = "https://" + url
        return url

    def copy(self):
        return CKANAPI(server=self.server, api_key=self.api_key,
                       ssl_verify=self.verify)

    def is_available(self):
        """Check whether server and API are reachable"""
        try:
            self.get("site_read")
        except BaseException:
            status = False
        else:
            status = True
        return status

    def get(self, api_call, **kwargs):
        """GET request

        Parameters
        ----------
        api_call: str
            An API call function (e.g. "package_show")
        kwargs: Any
            Any keyword arguments to the API call
            (e.g. `name="test-dataset"`)

        Returns
        -------
        result: dict
            Result of the API call converted to a dictionary
            from the returned json string
        """
        if kwargs:
            # Add keyword arguments
            kwv = []
            for kw in kwargs:
                kwv.append("{}={}".format(kw, kwargs[kw]))
            api_call += "?" + "&".join(kwv)
        url_call = self.api_url + api_call
        req = requests.get(url_call, headers=self.headers, verify=self.verify)
        try:
            rdata = req.json()
        except BaseException:
            rdata = {}
        if isinstance(rdata, str):
            raise ValueError(
                "Command did not succeed, output: '{}'".format(rdata))
        if not req.ok:
            error = rdata.get("error", {})
            etype = error.get("__type", req.reason)
            etext = ""
            for key in error:
                if not key.startswith("_"):
                    etext += "{}: {}".format(key, error[key])
            msg = "{}: {} (for '{}')".format(etype, etext, api_call)
            if req.reason == "NOT FOUND":
                raise APINotFoundError(msg)
            else:
                raise ConnectionError(msg)
        elif not rdata["success"]:
            raise ConnectionError(
                "Could not run API call '{}'! ".format(url_call)
                + "Reason: {} ({})".format(req.reason, rdata["error"]))
        return rdata["result"]

    def get_license_list(self):
        """Return the servers license list

        License lists are cached in :const:`SERVER_LICENCES`.
        """
        if self.api_url not in SERVER_LICENCES:
            SERVER_LICENCES[self.api_url] = self.get("license_list")
        return copy.deepcopy(SERVER_LICENCES[self.api_url])

    def get_supplementary_resource_schema(self):
        """Return the servers supplementary resource schema

        Schemas are cached in :const:`SERVER_RSS`.
        """
        if self.api_url not in SERVER_RSS:
            SERVER_RSS[self.api_url] = self.get("resource_schema_supplements")
        return copy.deepcopy(SERVER_RSS[self.api_url])

    def get_supported_resource_suffixes(self):
        """Return the servers supported resource suffixes

        Suffix lists are cached in :const:`SERVER_RSUFFIX`.
        """
        if self.api_url not in SERVER_RSUFFIX:
            SERVER_RSUFFIX[self.api_url] = self.get(
                "supported_resource_suffixes")
        return copy.deepcopy(SERVER_RSUFFIX[self.api_url])

    def get_user_dict(self):
        """Return the current user data dictionary

        The user name is inferred from the user list.
        """
        # Workaround for https://github.com/ckan/ckan/issues/5490
        # Get the user for which the email field is visible.
        data = self.get("user_list")
        for user in data:
            if user.get("email", ""):
                userdata = user
                break
        else:
            raise APIKeyError(
                "Could not determine user data. Please check API key.")
        return userdata

    def post(self, api_call, data, dump_json=True, headers={}):
        """POST request

        Parameters
        ----------
        api_call: str
            An API call function (e.g. "package_create")
        data: dict, MultipartEncoder, ...
            The data connected to the post request. For
            "package_create", this would be a dictionary
            with the dataset name, author, license, etc.
        dump_json: bool
            If True (default) dump `data` into a json string.
            If False, `data` is not touched.
        headers: dict
            Additional headers (updates `self.headers`) for the
            POST request (used for multipart uploads).

        Returns
        -------
        result: dict
            Result of the API call converted to a dictionary
            from the returned json string
        """
        new_headers = copy.deepcopy(self.headers)
        new_headers.update(headers)
        if dump_json:
            if "Content-Type" in headers:
                raise ValueError("Do not specify 'Content-Type' with "
                                 + "`dump_json=True`!")
            # This is necessary because we cannot otherwise
            # create packages with tags (list of dicts);
            # We have to json-dump the dict.
            new_headers["Content-Type"] = "application/json"
            data = json.dumps(data)

        url_call = self.api_url + api_call
        req = requests.post(url_call,
                            data=data,
                            headers=new_headers,
                            verify=self.verify)
        try:
            rdata = req.json()
        except BaseException:
            rdata = {}
        if isinstance(rdata, str):
            raise ValueError(
                "Command did not succeed, output: '{}'".format(rdata))
        if not req.ok:
            error = rdata.get("error", {})
            etype = error.get("__type", req.reason)
            etext = ""
            for key in error:
                if not key.startswith("_"):
                    etext += "{}: {}".format(key, error[key])
            msg = "{}: {} (for '{}')".format(etype, etext, api_call)
            if req.reason == "NOT FOUND":
                raise APINotFoundError(msg)
            elif req.reason == "CONFLICT":
                raise APIConflictError(msg)
            elif req.reason == "Gateway Time-out":
                raise APIGatewayTimeoutError(msg)
            elif req.reason == "Bad Gateway":
                raise APIBadGatewayError(msg)
            elif req.reason == "FORBIDDEN":
                raise APIAuthorizationError(msg)
            else:
                raise ConnectionError(msg)
        if not rdata["success"]:
            raise ConnectionError(
                "Could not run API call '{}'! ".format(url_call)
                + "Reason: {} ({})".format(req.reason, rdata["error"]))
        return rdata["result"]



import pathlib

from dcoraid.gui.wizard import access_token

datapath = pathlib.Path(__file__).parent / "data"


def test_get_hostname():
    """This test uses the figshare datasets on SERVER"""
    hostname = access_token.get_hostname(datapath / "cptamerica.dcor-access",
                                         "42")
    assert hostname == "dcor.example.com"


def test_get_api_key():
    api_key = access_token.get_api_key(datapath / "cptamerica.dcor-access",
                                       "42")
    assert api_key == "7c0c7203-4e25-4b14-a118-553c496a7a52"


if __name__ == "__main__":
    # Run all tests
    loc = locals()
    for key in list(loc.keys()):
        if key.startswith("test_") and hasattr(loc[key], "__call__"):
            loc[key]()



import warnings

from .core import DBInterrogator, DBModel
from .util import ttl_cache


class APIModel(DBModel):
    def __init__(self, api, *args, **kwargs):
        """A CKAN-API-based database model"""
        db = APIInterrogator(api=api)
        super(APIModel, self).__init__(interrogator=db, *args, **kwargs)


class APIInterrogator(DBInterrogator):
    def __init__(self, api):
        self.api = api.copy()
        super(APIInterrogator, self).__init__()

    @ttl_cache(seconds=5)
    def get_circles(self, mode="public"):
        """Return the list of DCOR Circles
        """
        if mode == "user":
            # Organizations the user is a member of
            data = self.api.get("organization_list_for_user",
                                permission="read")
        else:
            data = self.api.get("organization_list")
        return data

    @ttl_cache(seconds=5)
    def get_collections(self, mode="public"):
        """Return the list of DCOR Collections"""
        if mode == "user":
            data = self.api.get("group_list_authz", am_member=True)
        else:
            data = self.api.get("group_list")
        return data

    @ttl_cache(seconds=3600)
    def get_datasets_user_following(self):
        user_data = self.get_user_data()
        data = self.api.get("dataset_followee_list",
                            id=user_data["name"])
        return data

    @ttl_cache(seconds=3600)
    def get_datasets_user_owned(self):
        """Get all the user's datasets"""
        user_data = self.get_user_data()
        numd = user_data["number_created_packages"]
        if numd > 1000:
            raise NotImplementedError(
                "Reached hard limit of 1000 results! "
                + "Please ask someone to implement this with `start`.")
        data2 = self.api.get("package_search",
                             q="*:*",
                             fq="creator_user_id:{}".format(user_data["id"]),
                             rows=numd+1)
        if data2["count"] != numd:
            raise ValueError("Number of user datasets don't match!")

        return data2["results"]

    def get_datasets_user_shared(self):
        warnings.warn("`APIInterrogator.get_datasets_user_shared` "
                      + "not yet implemented!")
        return []

    def get_license_list(self):
        """Return the servers license list"""
        return self.api.get_license_list()

    def get_supplementary_resource_schema(self):
        """Return the servers supplementary resource schema"""
        return self.api.get_supplementary_resource_schema()

    def get_supported_resource_suffixes(self):
        """Return the servers supported resource suffixes"""
        return self.api.get_supported_resource_suffixes()

    @ttl_cache(seconds=3600)
    def get_user_data(self):
        """Return the current user data dictionary"""
        return self.api.get_user_dict()

    @ttl_cache(seconds=3600)
    def get_users(self, ret_fullnames=False):
        """Return the list of DCOR users"""
        data = self.api.get("user_list")
        user_list = []
        full_list = []
        for dd in data:
            user_list.append(dd["name"])
            if dd["fullname"]:
                full_list.append(dd["fullname"])
            else:
                full_list.append(dd["name"])
        if ret_fullnames:
            return user_list, full_list
        else:
            return user_list

    def search_dataset(self, query, circles, collections, mode="public"):
        # https://docs.ckan.org/en/latest/user-guide.html#search-in-detail
        solr_circles = ["organization:{}".format(ci) for ci in circles]
        solr_circle_query = " OR ".join(solr_circles)

        solr_collections = ["groups:{}".format(co) for co in collections]
        solr_collections_query = " OR ".join(solr_collections)

        data = self.api.get("package_search",
                            q=query,
                            include_private=(mode == "user"),
                            fq="({}) AND ({})".format(solr_circle_query,
                                                      solr_collections_query),
                            rows=100,
                            )
        return data["results"]

    @property
    def local_timestamp(self):
        """Local database date in seconds since epoch"""
        return self.remote_timestamp

    @property
    def local_version_score(self):
        """Local database version"""
        return self.remote_version_score



import shutil
import tempfile
import time

from PyQt5 import QtCore

from dcoraid.api import APIConflictError

import common


TMPDIR = tempfile.mkdtemp(prefix=time.strftime(
    "dcoraid_test_%H.%M_"))

pytest_plugins = ["pytest-qt"]


def pytest_configure(config):
    """This is ran before all tests"""
    # disable update checking
    QtCore.QCoreApplication.setOrganizationName("DCOR")
    QtCore.QCoreApplication.setOrganizationDomain("dcor.mpl.mpg.de")
    QtCore.QCoreApplication.setApplicationName("dcoraid")
    QtCore.QSettings.setDefaultFormat(QtCore.QSettings.IniFormat)
    settings = QtCore.QSettings()
    settings.setIniCodec("utf-8")
    settings.value("user scenario", "dcor-dev")
    settings.setValue("auth/server", "dcor-dev.mpl.mpg.de")
    settings.setValue("auth/api key", common.get_api_key())
    settings.sync()
    # set global temp directory
    tempfile.tempdir = TMPDIR


def pytest_unconfigure(config):
    """
    called before test process is exited.
    """
    # clear global temp directory
    shutil.rmtree(TMPDIR, ignore_errors=True)


def pytest_sessionstart(session):
    """
    Called after the Session object has been created and
    before performing collection and entering the run test loop.
    """
    api = common.get_api()
    api.get("group_show", id=common.COLLECTION)
    try:
        api.post("group_create", {"name": common.COLLECTION})
    except APIConflictError:
        pass
    try:
        api.post("organization_create", {"name": common.CIRCLE})
    except APIConflictError:
        pass
    user_dict = api.get("user_show", id=common.USER)
    user_dict["fullname"] = common.USER_NAME
    api.post("user_update", user_dict)



from collections import OrderedDict
import copy
from functools import lru_cache
import pathlib

from PyQt5 import QtCore, QtGui


class ResourcesModel(QtCore.QAbstractListModel):
    """Handle resources and their metadata selected in the UI

    This is the "Model" component of the QListView.
    """

    def __init__(self, *args, **kwargs):
        super(ResourcesModel, self).__init__(*args, **kwargs)
        self.resources = OrderedDict()

    def add_resources(self, rslist):
        """Add resources to the current model

        Note that resource suffixes are not checked for validity
        using the `supported_resource_suffixes` API command.
        This should be done in the UI file dialogs.
        """
        for ff in rslist:
            if ff not in self.resources:  # avoid adding the same file twice
                self.resources[ff] = {}
                self.layoutChanged.emit()

    def data(self, index, role=QtCore.Qt.DisplayRole):
        """Return data for 'View'"""
        if role == QtCore.Qt.DisplayRole:
            _, data = self.get_data_for_index(index)
            return data["file"]["filename"]

        if role == QtCore.Qt.DecorationRole:
            if self.index_has_edits(index):
                return get_icon("tag")
            else:
                return get_icon("slash")

    def filenames_are_unique(self):
        """Test whether the final file names are unique"""
        data = self.get_all_data()
        names = [data[dd]["file"]["filename"] for dd in data]
        return len(names) == len(list(set(names)))

    def filenames_were_edited(self):
        """Return number of filenames that have been edited"""
        data = self.get_all_data()
        counter = 0
        for key in data:
            if pathlib.Path(key).name != data[key]["file"]["filename"]:
                counter += 1
        return counter

    def index_is_dc(self, index):
        """Does the given index instance belong to an RT-DC file?"""
        path = pathlib.Path(self.get_file_list()[index.row()])
        if path.suffix in [".dc", ".rtdc"]:
            return True
        else:
            return False

    def index_has_edits(self, index):
        """Is there a modification of the list entry of this index instance?"""
        filen = self.get_file_list()[index.row()]
        return bool(self.resources[filen])

    def supplements_were_edited(self):
        """Return number of resource supplements that have been edited"""
        data = self.get_all_data()
        counter = 0
        for key in data:
            if data[key]["supplement"]:
                counter += 1
        return counter

    def get_all_data(self, magic_keys=True):
        """Return dictionary with complete information for all resources"""
        data = {}
        for ii, path in enumerate(self.resources.keys()):
            data[path] = self.get_data_for_row(ii, magic_keys=magic_keys)[1]
        return data

    def get_common_supplements_from_indexes(self, indexes):
        """Return the supplementary items common to indexes"""
        if indexes:
            common = self.get_data_for_index(indexes[0])[1]["supplement"]
            for idx in indexes:
                supi = self.get_data_for_index(idx)[1]["supplement"]
                for sec in list(common.keys()):
                    if sec not in supi:
                        common.pop(sec)
                    else:
                        for key in list(common[sec].keys()):
                            if (key not in supi[sec]
                                    or supi[sec][key] != common[sec][key]):
                                common[sec].pop(key)
        else:
            common = {}
        return common

    def get_data_for_index(self, index):
        """Return the complete resource dictionary for this index"""
        return self.get_data_for_row(index.row())

    def get_data_for_row(self, row, magic_keys=True):
        """Return the complete information dictionary for this row index"""
        rfile = self.get_file_list()[row]
        data = copy.deepcopy(self.resources[rfile])
        if "file" not in data:
            data["file"] = {}
        if "filename" not in data["file"]:
            data["file"]["filename"] = pathlib.Path(rfile).name
        if "supplement" not in data:
            data["supplement"] = {}
        if not magic_keys:
            for sec in list(data["supplement"].keys()):
                for key in list(data["supplement"][sec].keys()):
                    if key.startswith("MAGIC_"):
                        data["supplement"][sec].pop(key)
        return rfile, data

    def get_file_list(self):
        """Return the list of paths in the model"""
        return [item for item in self.resources]

    def get_filename_from_index(self, index):
        """Return the full path that belongs to a given index instance"""
        row = index.row()
        return self.get_file_list()[row]

    def get_indexes_types(self, indexes):
        """Return which types of files are selected

        Returns
        -------
        ftype: str
            None if selection is None, "dc" if selection is
            exclusively DC data, "mixed", if DC data and other
            resources are selected, "nodc" if no DC resources
            are selected.
        """
        dc_count = sum([self.index_is_dc(idx) for idx in indexes])
        if len(indexes) == 0:
            return None
        elif dc_count == len(indexes):
            return "dc"
        elif dc_count == 0:
            return "nodc"
        else:
            return "mixed"

    def rem_resources(self, indexes):
        """Remove resources in this list of index instances"""
        ids = [idx.row() for idx in indexes]
        for ii, pp in enumerate(self.get_file_list()):
            if ii in ids:
                self.resources.pop(pp)
        self.layoutChanged.emit()

    def rowCount(self, index=None):
        """Return number of resources"""
        return len(self.resources)

    @QtCore.pyqtSlot(list, dict)
    def update_resources(self, indexes, data_dict):
        """Update resources defined by index list with `data_dict`"""
        if len(indexes) == 0:  # nothing to do
            return
        elif len(indexes) == 1 and "file" in data_dict:
            # update the name
            pp = self.get_filename_from_index(indexes[0])
            self.resources[pp].update({"file": data_dict["file"]})

        if "supplement" in data_dict:
            for idx in indexes:
                if self.index_is_dc(idx):
                    # update supplementary parameters
                    pp = self.get_filename_from_index(idx)
                    self.resources[pp].update(
                        {"supplement": data_dict["supplement"]})

        self.layoutChanged.emit()


@lru_cache(maxsize=32)
def get_icon(name):
    return QtGui.QIcon.fromTheme(name).pixmap(16, 16)



from . import filter_base


class FilterCircles(filter_base.FilterBase):
    def __init__(self, *args, **kwargs):
        super(FilterCircles, self).__init__(*args, **kwargs)
        self.label.setText("Circles")
        self.lineEdit.setPlaceholderText("filter names...")
        self.checkBox.setVisible(False)


class FilterCollections(filter_base.FilterBase):
    def __init__(self, *args, **kwargs):
        super(FilterCollections, self).__init__(*args, **kwargs)
        self.label.setText("Collections")
        self.lineEdit.setPlaceholderText("filter names...")
        self.checkBox.setVisible(False)


class FilterDatasets(filter_base.FilterBase):
    def __init__(self, *args, **kwargs):
        super(FilterDatasets, self).__init__(*args, **kwargs)
        self.label.setText("Datasets")
        self.lineEdit.setPlaceholderText("filter titles...")
        self.checkBox.setVisible(False)


class FilterResources(filter_base.FilterBase):
    def __init__(self, *args, **kwargs):
        super(FilterResources, self).__init__(*args, **kwargs)
        self.label.setText("Resources")
        self.lineEdit.setPlaceholderText("filter file names...")
        self.checkBox.setVisible(True)
        self.checkBox.setText(".rtdc only")
        self.checkBox.setChecked(True)



from collections import OrderedDict
import pkg_resources
import warnings

from PyQt5 import QtCore, QtWidgets, uic


class FilterChain(QtWidgets.QWidget):
    def __init__(self, *args, **kwargs):
        """Filter chain widget with multiple filter views
        """
        super(FilterChain, self).__init__(*args, **kwargs)
        QtWidgets.QMainWindow.__init__(self)
        path_ui = pkg_resources.resource_filename("dcoraid.gui.dbview",
                                                  "filter_chain.ui")
        uic.loadUi(path_ui, self)

        #: Currently visible datasets
        self.datasets = []

        #: Signals and slots
        self.fw_datasets.selection_changed.connect(self.show_resources)
        self.fw_resources.checkBox.stateChanged.connect(self.update_resources)

    @property
    def selected_circles(self):
        """Circles currently selected"""
        return self.fw_circles.get_item_keys(selected=True)

    @property
    def selected_collections(self):
        """Collections currently selected"""
        return self.fw_collections.get_item_keys(selected=True)

    def set_db_extract(self, db_extract):
        """Set the database model

        Parameters
        ----------
        db_model: dcoraid.dbmodel.core.DBExtract
            Subclass of DBModel
        """
        self.db_extract = db_extract
        # reset all lists
        # circles
        circle_items = OrderedDict()
        for ci in self.db_extract.circles:
            warnings.warn("Have to implement global circle cache")
            circle_items[ci] = ci
        self.fw_circles.set_items(circle_items)
        # collections
        collection_items = OrderedDict()
        for co in self.db_extract.collections:
            warnings.warn("Have to implement global collection cache")
            collection_items[co] = co
        self.fw_collections.set_items(collection_items)
        # datasets
        dataset_items = OrderedDict()
        for ds in self.db_extract.datasets:
            dataset_items[ds["name"]] = ds["title"]
        self.fw_datasets.set_items(dataset_items)

    @QtCore.pyqtSlot(list)
    def show_resources(self, dataset_names):
        self.dataset_names = dataset_names
        self.update_resources()

    @QtCore.pyqtSlot()
    def update_resources(self):
        rs_items = OrderedDict()
        for dn in self.dataset_names:
            ddict = self.db_extract.get_dataset_dict(dn)
            for rs in ddict["resources"]:
                if (self.fw_resources.checkBox.isChecked()
                    and ("mimetype" in rs
                         and rs["mimetype"] != "RT-DC")):
                    # Ignore non-RT-DC mimetypes
                    continue
                else:
                    key = rs["id"]
                    name = rs["name"]
                    if "dc:experiment:event count" in rs:
                        name += " ({} events)".format(
                            rs["dc:experiment:event count"])
                    rs_items[key] = name
        self.fw_resources.set_items(rs_items)



import pkg_resources

from PyQt5 import uic, QtCore, QtWidgets


class TitleItem(QtWidgets.QWidget):
    def __init__(self, requires, title, *args, **kwargs):
        """A schema widget for the section titles

        Parameters
        ----------
        requires: dict
            Requirements dictionary
        title: str
            Section title
        """
        super(TitleItem, self).__init__(*args, **kwargs)

        self.verticalLayout = QtWidgets.QVBoxLayout(self)
        self.verticalLayout.setContentsMargins(0, 0, 0, 0)

        self.label = QtWidgets.QLabel("<b>{}</b>".format(title))
        self.verticalLayout.addWidget(self.label)

        self.requires = requires

    @QtCore.pyqtSlot(str, str, object, bool)
    def on_schema_key_toggled(self, section, key, value, enabled):
        """Hide or show this widget based on self.requires"""
        if not self.requires:
            self.setVisible(True)
        elif section in self.requires and key in self.requires[section]:
            if value in self.requires[section][key]:
                self.setVisible(enabled)
            else:
                self.setVisible(False)


class RSSItemBase(QtWidgets.QWidget):
    value_changed = QtCore.pyqtSignal(str, str, object, bool)

    def __init__(self, rss_dict, section, *args, **kwargs):
        super(RSSItemBase, self).__init__(*args, **kwargs)
        self.rss_dict = rss_dict
        self.section = section
        self.key = rss_dict["key"]
        self.requires = rss_dict.get("requires", {})

    def emit_value(self):
        """Emit `value_changed` with current data"""
        self.value_changed.emit(self.section, self.key, self.get_value(),
                                self.checkBox.isChecked())

    @QtCore.pyqtSlot()
    def on_value_changed(self):
        """Activate checkbox and call `emit_value`"""
        if self.sender() == self.checkBox and not self.checkBox.isChecked():
            # Do not check the checkBox again if the user unchecks it
            pass
        else:
            self.check(True)
        self.emit_value()

    @QtCore.pyqtSlot(str, str, object, bool)
    def on_schema_key_toggled(self, section, key, value, enabled):
        """Hide or show this widget based on self.rss_dict["requires"]"""
        if not self.requires:
            visible = True
        elif section in self.requires and key in self.requires[section]:
            if value in self.requires[section][key]:
                visible = enabled
            else:
                visible = False
        else:
            visible = self.isVisible()
        self.setVisible(visible)
        if not visible:
            # uncheck the widget if it is hidden
            self.checkBox.setChecked(False)


class RSSTagsItem(RSSItemBase):
    def __init__(self, *args, **kwargs):
        """Represents an item in the supplementary resource schema"""
        super(RSSTagsItem, self).__init__(*args, **kwargs)
        path_ui = pkg_resources.resource_filename(
            "dcoraid.gui.upload", "widget_supplement_tags.ui")
        uic.loadUi(path_ui, self)
        self.tableWidget.setRowCount(3)
        self.on_assert_row_count()

        # signals
        self.tableWidget.itemChanged.connect(self.on_value_changed)
        self.tableWidget.itemChanged.connect(self.on_assert_row_count)

    def check(self, b):
        """Check the check box"""
        self.checkBox.setChecked(b)

    def get_compounds(self):
        """Get the compounds labels (1st column)"""
        return " ".join([it[0] for it in self.get_value() if it]).strip()

    def get_labels(self):
        """Get the fluorescence labels (2nd column)"""
        return " ".join([it[1] for it in self.get_value() if it]).strip()

    def get_tags(self):
        """Return the tags (antibodies and fluorphores connected with dash)"""
        tags = []
        for it in self.get_value():
            if it[0] and it[1]:
                tags.append("{}-{}".format(*it))
        return " ".join(tags).strip()

    def get_value(self):
        """Return the items of the QTableWidget

        This is not used by DCOR, but only internally by DCOR-Aid.
        """
        self.on_assert_row_count()
        value = []
        for ii in range(self.tableWidget.rowCount()):
            it0 = self.tableWidget.item(ii, 0).text().strip()
            it1 = self.tableWidget.item(ii, 1).text().strip()
            if it0 or it1:
                value.append((it0, it1))
        return value

    def set_value(self, value):
        """Set the value"""
        self.on_assert_row_count()
        self.tableWidget.setRowCount(len(value))
        for ii, it in enumerate(value):
            self.tableWidget.item(ii, 0).setText(it[0])
            self.tableWidget.item(ii, 1).setText(it[1])

    @QtCore.pyqtSlot()
    def on_assert_row_count(self):
        """Increase row count by one if last item has data"""
        nrow = self.tableWidget.rowCount()
        it0 = self.tableWidget.item(nrow-1, 0)
        it1 = self.tableWidget.item(nrow-1, 1)
        if (it0 and it0.text().strip() or it1 and it1.text().strip()):
            nrow += 1
            self.tableWidget.setRowCount(nrow)
        # populate empty items
        for ii in range(nrow):
            for jj in [0, 1]:
                it = self.tableWidget.item(ii, jj)
                if it is None:
                    self.tableWidget.setItem(ii, jj,
                                             QtWidgets.QTableWidgetItem())


class RSSItem(RSSItemBase):
    def __init__(self, *args, **kwargs):
        """Represents an item in the supplementary resource schema"""
        super(RSSItem, self).__init__(*args, **kwargs)
        path_ui = pkg_resources.resource_filename(
            "dcoraid.gui.upload", "widget_supplement_item.ui")
        uic.loadUi(path_ui, self)
        self.apply_schema()

        # signals
        self.checkBox.clicked.connect(self.on_value_changed)
        self.spinBox.valueChanged.connect(self.on_value_changed)
        self.doubleSpinBox.valueChanged.connect(self.on_value_changed)
        self.lineEdit.textChanged.connect(self.on_value_changed)
        self.plainTextEdit.textChanged.connect(self.on_value_changed)
        self.comboBox.currentTextChanged.connect(self.on_value_changed)
        self.radioButton_yes.clicked.connect(self.on_value_changed)
        self.radioButton_no.clicked.connect(self.on_value_changed)

    def apply_schema(self):
        """Initialize the item schema according to self.rss_dict"""
        rss_dict = self.rss_dict
        self.checkBox.setText(rss_dict["name"])
        self.checkBox.setToolTip(rss_dict.get("hint", None))

        widget = self.get_data_widget()
        self.show_only_data_widget()
        if "choices" in rss_dict:  # combobox
            assert widget is self.comboBox
            self.comboBox.insertItems(0, rss_dict["choices"])
            editable = "choices fixed" not in rss_dict.get("options", [])
            self.comboBox.setEditable(editable)
            if editable:
                self.comboBox.setToolTip("You may also type your own choice.")
        else:
            example = rss_dict.get("example", None)
            if example and hasattr(widget, "setPlaceholderText"):
                widget.setPlaceholderText("e.g. {}".format(example))
        if "unit" in rss_dict:
            self.doubleSpinBox.setSuffix(" " + rss_dict["unit"])

    def check(self, b):
        """Check the check box"""
        self.checkBox.setChecked(b)

    def get_value(self):
        """Return the value of the current data widget"""
        _, value = self.get_data_widget(retval=True)
        return value

    def set_value(self, value):
        """Set the widget value with the appropriate function"""
        widget = self.get_data_widget()
        if widget is self.lineEdit:
            self.lineEdit.setText(value)
        elif widget is self.comboBox:
            self.comboBox.setCurrentText(value)
        elif widget is self.spinBox or widget is self.doubleSpinBox:
            widget.setValue(value)
        elif widget is self.plainTextEdit:
            widget.setPlainText(value)
        elif widget is self.widget_bool:
            if value:
                self.radioButton_yes.setChecked(True)
            else:
                self.radioButton_no.setChecked(True)

    def get_data_widget(self, retval=False):
        """Return the widget the holds the data according to self.rss_dict"""
        self.blockSignals(True)
        rss_dict = self.rss_dict
        if "choices" in rss_dict:  # combobox
            widget = self.comboBox
            value = self.comboBox.currentText()
        else:
            itemtype = rss_dict.get("type", "string")
            if itemtype == "string":
                if "text" in rss_dict.get("options", []):
                    widget = self.plainTextEdit
                    value = self.plainTextEdit.toPlainText()
                else:
                    widget = self.lineEdit
                    value = self.lineEdit.text()
            elif itemtype == "list":
                widget = self.lineEdit
                value = self.lineEdit.text()
            elif itemtype == "float":
                widget = self.doubleSpinBox
                value = self.doubleSpinBox.value()
            elif itemtype == "integer":
                widget = self.spinBox
                value = self.spinBox.value()
            elif itemtype == "boolean":
                widget = self.widget_bool
                value = self.radioButton_yes.isChecked()
            else:
                raise ValueError("No rule to process item {}".format(rss_dict))
        self.blockSignals(False)
        if retval:
            return widget, value
        else:
            return widget

    def show_only_data_widget(self):
        """Convenience function that hides all but the data widget"""
        widget = self.get_data_widget()
        for ww in [self.comboBox,
                   self.doubleSpinBox,
                   self.lineEdit,
                   self.plainTextEdit,
                   self.spinBox,
                   self.widget_bool]:
            if ww is not widget:
                ww.hide()



import pkg_resources
import sys
import uuid

from PyQt5 import uic, QtCore, QtWidgets

from ...api import APIKeyError, CKANAPI

from .import access_token


def get_dcor_dev_api_key():
    """Return a valid DCOR-dev user API key

    If there are no credentials available in the settings file,
    then a new user is created on DCOR-dev.
    """
    settings = QtCore.QSettings()
    api_key = settings.value("auth/api key", "")
    api = CKANAPI(server="https://dcor-dev.mpl.mpg.de",
                  api_key=api_key,
                  ssl_verify=True)
    try:
        api.get_user_dict()
    except APIKeyError:
        # create a new user
        rstr = str(uuid.uuid4())
        pwd = str(uuid.uuid4())[:8]
        usr = "dcoraid-{}".format(rstr[:5])
        user_dict = api.post(
            "user_create",
            data={"name": usr,
                  "fullname": "Player {}".format(rstr[:5]),
                  "email": "{}@dcor-dev.mpl.mpg.de".format(usr),
                  "password": pwd,
                  })
        api_key = user_dict["apikey"]
        # Let the user know that he has a password
        QtWidgets.QMessageBox.information(
            None, "You have a password",
            "DCOR-Aid generated a password for you which you could use "
            + "to view your private datasets online. If you would like "
            + "to test private datasets, please record this: "
            + "\n\n user: {}\n password: {}".format(usr, pwd)
            + "\n\nAlternatively, you could at any point set a valid "
            + "email address in the user preferences and request a "
            + "password reset link via the web interface.")
    return api_key


class SetupWizard(QtWidgets.QWizard):
    """DCOR-Aid setup wizard"""

    def __init__(self, *args, **kwargs):
        super(SetupWizard, self).__init__(None)
        path_ui = pkg_resources.resource_filename(
            "dcoraid.gui.wizard", "wizard.ui")
        uic.loadUi(path_ui, self)

        self.pushButton_path_access_token.clicked.connect(
            self.on_browse_access_token)
        self.button(QtWidgets.QWizard.FinishButton).clicked.connect(
            self._finalize)

    def _finalize(self):
        """Update settings"""
        user_scenario = self.get_user_scenario()
        server = self.get_server()
        api_key = self.get_api_key()
        settings = QtCore.QSettings()
        old_server = settings.value("auth/server", "")
        old_api_key = settings.value("auth/api key", "")
        old_user_scenario = settings.value("user scenario", "")
        if (user_scenario != old_user_scenario
            or old_api_key != api_key
                or old_server != server):
            if old_api_key:
                msg = "Changing the server or API key requires a restart of " \
                      + "DCOR-Aid. If you choose 'No', then the original " \
                      + "server and API key are NOT changed. Do you really " \
                      + "want to quit DCOR-Aid?"
                button_reply = QtWidgets.QMessageBox.question(
                    self,
                    'DCOR-Aid restart required',
                    msg,
                    QtWidgets.QMessageBox.Yes | QtWidgets.QMessageBox.No,
                    QtWidgets.QMessageBox.No)
                if button_reply == QtWidgets.QMessageBox.Yes:
                    proceed = True
                else:
                    proceed = False
            else:
                QtWidgets.QMessageBox.information(
                    self,
                    "DCOR-Aid restart required",
                    "Please restart DCOR-Aid to proceed."
                )
                proceed = True

            if proceed:
                settings.setValue("user scenario",
                                  self.get_user_scenario())
                settings.setValue("auth/server", server)
                settings.setValue("auth/api key", api_key)
                if user_scenario == "medical":
                    # save the server certificate
                    cert_data = access_token.get_certificate(
                        self.lineEdit_access_token_path.text(),
                        self.lineEdit_access_token_password.text())
                    settings.setValue("auth/certificate", cert_data)
                elif settings.contains("auth/certificate"):
                    settings.remove("auth/certificate")
                settings.sync()
                QtWidgets.QApplication.processEvents(
                    QtCore.QEventLoop.AllEvents, 300)
                QtWidgets.QApplication.quit()
                sys.exit(0)  # if the above does not work

    def get_api_key(self):
        user_scenario = self.get_user_scenario()
        if user_scenario == "dcor":
            return self.lineEdit_apikey_dcor.text()
        elif user_scenario == "medical":
            return access_token.get_api_key(
                self.lineEdit_access_token_path.text(),
                self.lineEdit_access_token_password.text())
        elif user_scenario == "dcor-dev":
            return get_dcor_dev_api_key()
        elif user_scenario == "anonymous":
            return ""
        elif user_scenario == "custom":
            return self.lineEdit_apikey_custom.text()
        else:
            raise ValueError("Undefined!")

    def get_server(self):
        user_scenario = self.get_user_scenario()
        if user_scenario == "dcor":
            return "dcor.mpl.mpg.de"
        elif user_scenario == "medical":
            return access_token.get_hostname(
                self.lineEdit_access_token_path.text(),
                self.lineEdit_access_token_password.text())
        elif user_scenario == "dcor-dev":
            return "dcor-dev.mpl.mpg.de"
        elif user_scenario == "anonymous":
            return "dcor.mpl.mpg.de"
        elif user_scenario == "custom":
            return self.lineEdit_server_custom.text()
        else:
            raise ValueError("Undefined!")

    def get_user_scenario(self):
        """Return string representing the user scenario (from first page)"""
        if self.radioButton_dcor.isChecked():
            return "dcor"
        elif self.radioButton_med.isChecked():
            return "medical"
        elif self.radioButton_play.isChecked():
            return "dcor-dev"
        elif self.radioButton_anonym.isChecked():
            return "anonymous"
        elif self.radioButton_custom.isChecked():
            return "custom"
        else:
            raise ValueError("Undefined!")

    def nextId(self):
        """Determine the next page based on the current page data"""
        user_scenario = self.get_user_scenario()
        page = self.currentPage()
        page_dict = {}
        for ii in self.pageIds():
            page_dict[self.page(ii)] = ii
        if page == self.page_welcome:
            if user_scenario == "dcor":
                return page_dict[self.page_dcor]
            elif user_scenario == "medical":
                return page_dict[self.page_med]
            elif user_scenario == "dcor-dev":
                return -1
            elif user_scenario == "anonymous":
                return -1
            elif user_scenario == "custom":
                return page_dict[self.page_custom]
            else:
                raise ValueError("No Rule!")
        else:
            return -1

    @QtCore.pyqtSlot()
    def on_browse_access_token(self):
        path, _ = QtWidgets.QFileDialog.getOpenFileName(
            self, 'Open access token', '', 'DCOR access token (*.dcor-access)',
            'DCOR access token (*.dcor-access)')
        if path:
            self.lineEdit_access_token_path.setText(path)



import pkg_resources

from PyQt5 import uic, QtCore, QtWidgets

from ...upload.dataset import remove_all_drafts

from ..api import get_ckan_api
from ..tools import ShowWaitCursor


class MaintenanceWidget(QtWidgets.QWidget):
    def __init__(self, *args, **kwargs):
        """Maintenance tasks
        """
        super(MaintenanceWidget, self).__init__(*args, **kwargs)
        path_ui = pkg_resources.resource_filename(
            "dcoraid.gui.maintenance", "widget_maintenance.ui")
        uic.loadUi(path_ui, self)

        self.toolButton_drafts.clicked.connect(self.on_remove_drafts)

    @QtCore.pyqtSlot()
    def on_remove_drafts(self):
        with ShowWaitCursor():
            data = remove_all_drafts(api=get_ckan_api())
        msg = QtWidgets.QMessageBox()
        msg.setIcon(QtWidgets.QMessageBox.Information)
        if len(data):
            details = ["{} ({})".format(d["title"], d["name"]) for d in data]
            msg.setText("Drafts removed: {}".format(len(data)))
            msg.setDetailedText("\n".join(details))
            msg.setWindowTitle("Success")
        else:
            msg.setText("No drafts found.")
            msg.setWindowTitle("Nothing to do")
        msg.exec_()



import os
import pathlib

from dcoraid.api import CKANAPI


CIRCLE = "dcoraid-circle"
COLLECTION = "dcoraid-collection"
DATASET = "dcoraid-dataset"
SERVER = "dcor-dev.mpl.mpg.de"
USER = "dcoraid"
USER_NAME = "DCOR-Aid tester"
TITLE = "DCOR-Aid test dataset"


def get_api():
    api = CKANAPI(server=SERVER, api_key=get_api_key(), ssl_verify=True)
    return api


def get_api_key():
    key = os.environ.get("DCOR_API_KEY")
    if not key:
        # local file
        kp = pathlib.Path(__file__).parent / "api_key"
        if not kp.exists():
            raise ValueError("No DCOR_API_KEY variable or api_key file!")
        key = kp.read_text().strip()
    return key



import os
import pathlib
import time

from dcoraid.upload import UploadQueue, create_dataset

import common


dpath = pathlib.Path(__file__).parent / "data" / "calibration_beads_47.rtdc"


def get_api_key():
    key = os.environ.get("DCOR_API_KEY")
    if not key:
        # local file
        kp = pathlib.Path(__file__).parent / "api_key"
        if not kp.exists():
            raise ValueError("No DCOR_API_KEY variable or api_key file!")
        key = kp.read_text().strip()
    return key


def make_dataset_dict(hint=""):
    hint += " "
    dataset_dict = {
        "title": "A test dataset {}{}".format(hint, time.time()),
        "private": True,
        "license_id": "CC0-1.0",
        "owner_org": common.CIRCLE,
        "authors": common.USER_NAME,
    }
    return dataset_dict


def test_queue_create_dataset_with_resource():
    api = common.get_api()
    # create some metadata
    dataset_dict = make_dataset_dict(hint="create-with-resource")
    # post dataset creation request
    data = create_dataset(dataset_dict=dataset_dict, api=api)
    joblist = UploadQueue(api=api)
    joblist.add_job(dataset_dict=data,
                    paths=[dpath])
    for _ in range(600):  # 60 seconds to upload
        if joblist[0].state == "done":
            break
        time.sleep(.1)
    else:
        assert False, "Job not finished: {}".format(joblist[0].get_status())


if __name__ == "__main__":
    # Run all tests
    loc = locals()
    for key in list(loc.keys()):
        if key.startswith("test_") and hasattr(loc[key], "__call__"):
            loc[key]()



import pathlib
import pkg_resources

from PyQt5 import uic, QtCore, QtWidgets

from ...upload import UploadQueue

from ..api import get_ckan_api
from ..tools import ShowWaitCursor

from .dlg_upload import UploadDialog
from .widget_tablecell_actions import TableCellActions


class UploadWidget(QtWidgets.QWidget):
    upload_finished = QtCore.pyqtSignal()

    def __init__(self, *args, **kwargs):
        """Manage running uploads
        """
        super(UploadWidget, self).__init__(*args, **kwargs)
        path_ui = pkg_resources.resource_filename(
            "dcoraid.gui.upload", "widget_upload.ui")
        uic.loadUi(path_ui, self)

        self.toolButton_new_upload.clicked.connect(self.on_draft_upload)

        # Underlying upload class
        self.jobs = UploadQueue(api=get_ckan_api())
        self.widget_jobs.set_job_list(self.jobs)

        # upload finished signal
        self.widget_jobs.upload_finished.connect(self.upload_finished)

    @QtCore.pyqtSlot()
    def on_draft_upload(self):
        """Guides the user through the process of creating a dataset

        A draft dataset is created to which the resources are then
        uploaded.
        """
        with ShowWaitCursor():
            dlg = UploadDialog(self)
            dlg.finished.connect(self.on_run_upload)
        dlg.exec()

    @QtCore.pyqtSlot(object)
    def on_run_upload(self, upload_dialog):
        """Proceed with resource upload as defined by `update_dialog`

        `update_dialog` is an instance of `dlg_upload.UploadDialog`
        and contains all information necessary to run the resource
        upload. Supplementary resource metadata is extracted here
        as well.
        """
        # Remove all magic keys from the schema data (they are
        # only used internally by DCOR-Aid and don't belong on DCOR).
        rdata = upload_dialog.rvmodel.get_all_data(magic_keys=False)
        paths = []
        names = []
        supps = []
        for path in rdata:
            paths.append(pathlib.Path(path))
            names.append(rdata[path]["file"]["filename"])
            supps.append(rdata[path]["supplement"])
        dataset_dict = upload_dialog.dataset_dict
        # add the entry to the job list
        self.jobs.add_job(dataset_dict, paths=paths, resource_names=names,
                          supplements=supps)


class UploadTableWidget(QtWidgets.QTableWidget):
    upload_finished = QtCore.pyqtSignal()

    def __init__(self, *args, **kwargs):
        super(UploadTableWidget, self).__init__(*args, **kwargs)
        self.jobs = []  # Will become UploadJobList with self.set_job_list
        self.timer = QtCore.QTimer()
        self.timer.timeout.connect(self.update_job_status)
        self.timer.start(30)
        self._finished_uploads = []

    def set_job_list(self, jobs):
        """Set the current job list

        The job list can be a `list`, but it is actually
        an `UploadJobList`.
        """
        # This is the actual initialization
        self.jobs = jobs

    def on_job_abort(self, dataset_id):
        self.jobs.abort_job(dataset_id)

    def on_job_delete(self, dataset_id):
        self.jobs.remove_job(dataset_id)

    def on_upload_finished(self, dataset_id):
        """Triggers upload_finished whenever an upload is finished"""
        if dataset_id not in self._finished_uploads:
            self._finished_uploads.append(dataset_id)
            self.upload_finished.emit()

    @QtCore.pyqtSlot()
    def update_job_status(self):
        """Update UI with information from self.jobs (UploadJobList)"""
        # disable updates
        self.setUpdatesEnabled(False)
        # make sure the length of the table is long enough
        self.setRowCount(len(self.jobs))
        self.setColumnCount(6)

        for row, job in enumerate(self.jobs):
            status = job.get_status()
            self.set_label_item(row, 0, job.dataset_id[:5])
            self.set_label_item(row, 1, job.dataset_dict["title"])
            self.set_label_item(row, 2, status["state"])
            self.set_label_item(row, 3, job.get_progress_string())
            self.set_label_item(row, 4, job.get_rate_string())
            if status["state"] == "done":
                self.on_upload_finished(job.dataset_id)
            self.set_actions_item(row, 5, job)

        # spacing (did not work in __init__)
        header = self.horizontalHeader()
        header.setSectionResizeMode(
            0, QtWidgets.QHeaderView.ResizeToContents)
        header.setSectionResizeMode(1, QtWidgets.QHeaderView.Stretch)
        # enable updates again
        self.setUpdatesEnabled(True)

    def set_label_item(self, row, col, label):
        """Get/Create a Qlabel at the specified position

        User has to make sure that row and column count are set
        """
        label = "{}".format(label)
        item = self.item(row, col)
        if item is None:
            item = QtWidgets.QTableWidgetItem(label)
            item.setFlags(QtCore.Qt.ItemIsEnabled)
            self.setItem(row, col, item)
        else:
            if item.text() != label:
                item.setText(label)

    def set_actions_item(self, row, col, job):
        """Set/Create a TableCellActions widget in the table

        Refreshes the widget and also connects signals.
        """
        wid = self.cellWidget(row, col)
        if wid is None:
            wid = TableCellActions(job)
            wid.delete_job.connect(self.on_job_delete)
            wid.abort_job.connect(self.on_job_abort)
            self.setCellWidget(row, col, wid)
        wid.refresh_visibility(job)



import pathlib
import random

from dcoraid.upload import create_dataset
from dcoraid.dbmodel import model_api

import common

dpath = pathlib.Path(__file__).parent / "data" / "calibration_beads_47.rtdc"


def test_dcor_dev_public_api_interrogator():
    """This test uses the figshare datasets on SERVER"""
    api = common.get_api()
    db = model_api.APIInterrogator(api=api)
    assert common.CIRCLE in db.get_circles()
    assert common.COLLECTION in db.get_collections()
    assert common.USER in db.get_users()


def test_dcor_dev_user_data():
    """Test the user information"""
    api = common.get_api()
    db = model_api.APIInterrogator(api=api)
    data = db.get_user_data()
    assert data["fullname"] == common.USER_NAME, "fullname not correct"


def test_dcor_dev_search():
    api = common.get_api()
    ranstr = ''.join(random.choice("0123456789") for _i in range(10))
    # Create a test dataset
    create_dataset({"title": "{} {}".format(common.TITLE, ranstr),
                    "owner_org": common.CIRCLE,
                    "authors": common.USER_NAME,
                    "license_id": "CC0-1.0",
                    "groups": [{"name": common.COLLECTION}],
                    },
                   api=api,
                   resources=[dpath],
                   activate=True)
    db = model_api.APIInterrogator(api=api)
    # Positive test
    data = db.search_dataset(query="dcoraid",
                             circles=[common.CIRCLE],
                             collections=[common.COLLECTION],
                             )
    assert len(data) >= 1
    for dd in data:
        if dd["name"].endswith(ranstr):
            break
    else:
        assert False, "{} not found!".format(common.DATASET)
    # Negative test
    data = db.search_dataset(query="cliauwenlc_should_never_exist",
                             circles=[common.CIRCLE],
                             collections=[common.COLLECTION],
                             )
    assert len(data) == 0, "search result for non-existent dataset?"


if __name__ == "__main__":
    # Run all tests
    loc = locals()
    for key in list(loc.keys()):
        if key.startswith("test_") and hasattr(loc[key], "__call__"):
            loc[key]()



"""https://github.com/munshigroup/kthread"""
import ctypes
import inspect
import threading
import time


def _async_raise(tid, exctype):
    """Raises the exception, causing the thread to exit"""
    if not inspect.isclass(exctype):
        raise TypeError("Only types can be raised (not instances)")
    res = ctypes.pythonapi.PyThreadState_SetAsyncExc(
        ctypes.c_long(tid), ctypes.py_object(exctype))
    if res == 0:
        raise ValueError("Invalid thread ID")
    elif res != 1:
        # """if it returns a number greater than one, you're in trouble,
        # and you should call it again with exc=NULL to revert the effect"""
        ctypes.pythonapi.PyThreadState_SetAsyncExc(tid, 0)
        raise SystemError("PyThreadState_SetAsyncExc failed")


class KThread(threading.Thread):
    """Killable thread. See terminate() for details."""

    def _get_my_tid(self):
        """Determines the instance's thread ID"""
        if not self.is_alive():
            raise threading.ThreadError("Thread is not active")

        # do we have it cached?
        if hasattr(self, "_thread_id"):
            return self._thread_id

        # no, look for it in the _active dict
        for tid, tobj in threading._active.items():
            if tobj is self:
                self._thread_id = tid
                return tid

        raise AssertionError("Could not determine the thread's ID")

    def raise_exc(self, exctype):
        """raises the given exception type in the context of this thread"""
        _async_raise(self._get_my_tid(), exctype)

    def terminate(self):
        """raises SystemExit in the context of the given thread, which should
        cause the thread to exit silently (unless caught)"""
        # WARNING: using terminate() can introduce instability in your
        # programs. It is worth noting that terminate() will NOT work if the
        # thread in question is blocked by a syscall (accept(), recv(), etc.).
        while self.is_alive():
            self.raise_exc(SystemExit)
            time.sleep(0.01)



import pkg_resources
import random
import socket
import string
import traceback as tb

from PyQt5 import uic, QtCore, QtWidgets

from ...api import APIKeyError, CKANAPI
from ..tools import show_wait_cursor

from ..api import get_ckan_api


class PreferencesDialog(QtWidgets.QMainWindow):
    show_server = QtCore.pyqtSignal()
    show_user = QtCore.pyqtSignal()

    def __init__(self, parent=None):
        """Create a new window for preferences
        """
        super(PreferencesDialog, self).__init__(parent)
        path_ui = pkg_resources.resource_filename(
            "dcoraid.gui.preferences", "dlg_preferences.ui")
        uic.loadUi(path_ui, self)

        self.setWindowTitle("DCOR-Aid Preferences")
        self.show_server.connect(self.on_show_server)
        self.show_user.connect(self.on_show_user)
        self.tabWidget.currentChanged.connect(self.on_tab_changed)
        self.toolButton_user_update.clicked.connect(self.on_update_user)
        self.toolButton_server_update.clicked.connect(self.on_update_server)
        self.toolButton_api_token_renew.clicked.connect(
            self.on_api_token_renew)
        self.toolButton_api_token_revoke.clicked.connect(
            self.on_api_token_revoke)
        self.toolButton_eye.clicked.connect(self.on_toggle_api_password_view)

        self.settings = QtCore.QSettings()
        # hidden initially
        self.hide()

    def ask_change_server_or_api_key(self):
        """Ask user whether he really wants to change things

        ...because it implies a restart of DCOR-Aid.
        """
        button_reply = QtWidgets.QMessageBox.question(
            self,
            'DCOR-Aid restart required',
            "Changing the server or API token requires a restart of "
            + "DCOR-Aid. If you choose 'No', then the original server "
            + "and API key are NOT changed. Do you really want to quit "
            + "DCOR-Aid?",
            QtWidgets.QMessageBox.Yes | QtWidgets.QMessageBox.No,
            QtWidgets.QMessageBox.No)
        if button_reply == QtWidgets.QMessageBox.Yes:
            return True
        else:
            return False

    @QtCore.pyqtSlot()
    def on_toggle_api_password_view(self):
        cur_em = self.lineEdit_api_key.echoMode()
        if cur_em == QtWidgets.QLineEdit.Normal:
            new_em = QtWidgets.QLineEdit.PasswordEchoOnEdit
        else:
            new_em = QtWidgets.QLineEdit.Normal
        self.lineEdit_api_key.setEchoMode(new_em)

    @QtCore.pyqtSlot()
    def on_api_token_renew(self):
        if self.ask_change_server_or_api_key():
            api_key = self.settings.value("auth/api key")
            if len(api_key) == 36:
                # deprecated API key
                ret = QtWidgets.QMessageBox.question(
                    self,
                    "Deprecated API key",
                    "You are using an API key instead of an API token. "
                    + "API keys are deprecated and cannot be invalidated. "
                    + "DCOR-Aid can only remove it locally. A new API token "
                    + "will be created. Proceed?"
                )
                if ret != QtWidgets.QMessageBox.Yes:
                    # Abort
                    return
            # Create a new token
            api = get_ckan_api()
            # create a new token
            user_dict = api.get_user_dict()
            token_name = "DCOR-Aid {} {}".format(
                socket.gethostname(),
                ''.join(random.choice(string.ascii_letters) for _ in range(5)))

            tret = api.post("api_token_create",
                            data={"user": user_dict["id"],
                                  "name": token_name})
            self.settings.setValue("auth/api key", tret["token"])

            if len(api_key) != 36:
                # revoke the old API token
                api.post("api_token_revoke",
                         data={"token": api_key})
            QtWidgets.QApplication.quit()

    @QtCore.pyqtSlot()
    def on_api_token_revoke(self):
        if self.ask_change_server_or_api_key():
            api_key = self.settings.value("auth/api key")
            if len(api_key) == 36:
                # deprecated API key
                ret = QtWidgets.QMessageBox.question(
                    self,
                    "Deprecated API key",
                    "You are using an API key instead of an API token. "
                    + "API keys are deprecated and cannot be invalidated. "
                    + "DCOR-Aid can only remove it locally. Proceed?"
                )
                if ret != QtWidgets.QMessageBox.Yes:
                    # Abort
                    return
            else:
                # API token
                api = get_ckan_api()
                api.post("api_token_revoke",
                         data={"token": self.settings.value("auth/api key")})
            self.settings.remove("auth/api key")
            QtWidgets.QApplication.quit()

    @QtCore.pyqtSlot()
    def on_show_server(self):
        self.comboBox_server.clear()
        for server in self.settings.value("server list",
                                          ["dcor.mpl.mpg.de"]):
            self.comboBox_server.addItem(server)
        self.comboBox_server.setCurrentText(
            self.settings.value("auth/server", "dcor.mpl.mpg.de"))
        self.lineEdit_api_key.setText(self.settings.value("auth/api key", ""))
        self.tabWidget.setCurrentIndex(0)  # server settings
        self.lineEdit_api_key.setEchoMode(
            QtWidgets.QLineEdit.PasswordEchoOnEdit)
        self.show()
        self.activateWindow()

    @QtCore.pyqtSlot()
    @show_wait_cursor
    def on_show_user(self):
        api = get_ckan_api()
        try:
            user_dict = api.get_user_dict()
        except (ConnectionError, APIKeyError):
            msg = QtWidgets.QMessageBox()
            msg.setIcon(QtWidgets.QMessageBox.Warning)
            msg.setText("No connection or wrong server or invalid API key!")
            msg.setWindowTitle("Warning")
            msg.setDetailedText(tb.format_exc())
            msg.exec_()
            self.on_show_server()
        else:
            self.lineEdit_user_id.setText(user_dict["name"])
            self.lineEdit_user_name.setText(user_dict["fullname"])
            self.lineEdit_user_email.setText(user_dict["email"])
            self.plainTextEdit_user_about.setPlainText(user_dict["about"])
            self.show()
            self.activateWindow()

    @QtCore.pyqtSlot(int)
    def on_tab_changed(self, index):
        if index == 0:
            self.on_show_server()
        elif index == 1:
            self.on_show_user()

    @QtCore.pyqtSlot()
    @show_wait_cursor
    def on_update_user(self):
        api = get_ckan_api()
        try:
            user_dict = api.get_user_dict()
        except (ConnectionError, APIKeyError):
            msg = QtWidgets.QMessageBox()
            msg.setIcon(QtWidgets.QMessageBox.Warning)
            msg.setText("No connection or wrong server or invalid API key!")
            msg.setWindowTitle("Warning")
            msg.setDetailedText(tb.format_exc())
            msg.exec_()
            self.on_show_server()
        update_dict = {}
        update_dict["id"] = user_dict["id"]
        update_dict["fullname"] = self.lineEdit_user_name.text()
        update_dict["email"] = self.lineEdit_user_email.text()
        update_dict["about"] = self.plainTextEdit_user_about.toPlainText()
        api.post("user_update", data=update_dict)

    @QtCore.pyqtSlot()
    @show_wait_cursor
    def on_update_server(self):
        old_server = self.settings.value("auth/server", "")
        old_api_key = self.settings.value("auth/api key", "")
        api_key = self.lineEdit_api_key.text()
        if len(api_key) == 36:
            # deprecated API Key (UUID)
            valid = "0123456789abcdef-"
        else:
            # new API tokens
            valid = "0123456789" \
                    + "abcdefghijklmnopqrstuvwxyz" \
                    + "ABCDEFGHIJKLMNOPQRSTUVWXYZ" \
                    + "._-"
        api_key = "".join([ch for ch in api_key if ch in valid])
        server = self.comboBox_server.currentText().strip()
        # Test whether that works
        try:
            cur_api = get_ckan_api()  # maybe only name or api key changed
            api = CKANAPI(server=server, api_key=api_key,
                          ssl_verify=cur_api.verify)
            api.get_user_dict()  # raises an exception if credentials are wrong
        except BaseException:
            msg = QtWidgets.QMessageBox()
            msg.setIcon(QtWidgets.QMessageBox.Critical)
            msg.setText("Bad server / API key combination!")
            msg.setWindowTitle("Error")
            msg.setDetailedText(tb.format_exc())
            msg.exec_()
        else:
            if old_server != server or old_api_key != api_key:
                if self.ask_change_server_or_api_key():
                    self.settings.setValue("auth/api key", api_key)
                    self.settings.setValue("auth/server", server)
                    QtWidgets.QApplication.quit()



import json
import pathlib

from PyQt5.QtCore import QStandardPaths


class PersistentResourceSchemaPresets(object):
    def __init__(self):
        """Dict-like interface for schema presets for resource uploads"""
        self._presets = {}
        self.path = pathlib.Path(QStandardPaths.writableLocation(
            QStandardPaths.AppDataLocation)) / "upload_resource_schema_presets"
        self.path.mkdir(exist_ok=True, parents=True)
        for pp in self.path.glob("*.json"):
            with pp.open() as fd:
                self._presets[pp.stem] = json.load(fd)

    def __contains__(self, key):
        return self._presets.__contains__(key)

    def __getitem__(self, key):
        return self._presets[key]

    def __setitem__(self, key, value):
        self._presets[key] = value
        with (self.path / "{}.json".format(key)).open("w") as fd:
            json.dump(value, fd, indent=2)

    def keys(self):
        return self._presets.keys()



import time
import warnings

from .job import UploadJob
from .kthread import KThread


class UploadQueue(object):
    def __init__(self, api):
        self.api = api.copy()
        if not api.api_key:
            warnings.warn("No API key is set! Upload will not work!")
        self.jobs = []
        self.daemon_compress = CompressDaemon(self.jobs)
        self.daemon_upload = UploadDaemon(self.jobs)
        self.daemon_verify = VerifyDaemon(self.jobs)

    def __getitem__(self, index):
        return self.jobs[index]

    def __len__(self):
        return len(self.jobs)

    def abort_job(self, dataset_id):
        """Abort a running job but don't remove it from the queue"""
        job = self.get_job(dataset_id)
        if job.state == "transfer":
            job.set_state("abort")
            # https://github.com/requests/toolbelt/issues/297
            self.daemon_upload.terminate()
            self.daemon_upload = UploadDaemon(self.jobs)

    def add_job(self, dataset_dict, paths, resource_names=None,
                supplements=None):
        """Add a job to the job list"""
        job = UploadJob(dataset_dict=dataset_dict,
                        paths=paths,
                        resource_names=resource_names,
                        supplements=supplements,
                        api=self.api)
        self.jobs.append(job)

    def get_job(self, dataset_id):
        """Return the job instance belonging to the dataset ID"""
        for job in self.jobs:
            if job.dataset_id == dataset_id:
                return job
        else:
            raise KeyError("Job '{}' not found!".format(dataset_id))

    def get_status(self, dataset_id):
        """Return the status of an upload job"""
        self.get_job(dataset_id).get_status()

    def remove_job(self, dataset_id):
        """Remove a job from the queue and perform cleanup

        It has not been tested what happens when a running job
        is aborted. It will probably keep running and then complain
        about resources that are expected to be there. Don't do it.
        """
        for ii, job in enumerate(list(self.jobs)):
            if job.dataset_id == dataset_id:
                self.jobs.pop(ii)
                job.cleanup()


class Daemon(KThread):
    def __init__(self, queue, job_trigger_state, job_function_name):
        """Daemon base class"""
        self.queue = queue
        self.state = "running"
        self.job_trigger_state = job_trigger_state
        self.job_function_name = job_function_name
        super(Daemon, self).__init__()
        self.daemon = True  # We don't have to worry about ending this thread
        self.start()

    def join(self, *args, **kwargs):
        """Join thread by breaking the while loop"""
        self.state = "exiting"
        super(Daemon, self).join(*args, **kwargs)
        assert self.state == "exited"

    def run(self):
        while True:
            if self.state == "exiting":
                self.state = "exited"
                break
            elif self.state != "running":
                # Don't do anything
                time.sleep(.1)
                continue
            else:
                # Get the first job that is in the trigger state
                for job in self.queue:
                    if job.state == self.job_trigger_state:
                        break
                else:
                    # Nothing to do, sleep a little to avoid 100% CPU
                    time.sleep(.1)
                    continue
                # Perform daemon task
                task = getattr(job, self.job_function_name)
                task()


class CompressDaemon(Daemon):
    def __init__(self, jobs):
        """Compression daemon"""
        super(CompressDaemon, self).__init__(
            jobs,
            job_trigger_state="init",
            job_function_name="task_compress_resources")


class UploadDaemon(Daemon):
    def __init__(self, jobs):
        """Upload daemon"""
        super(UploadDaemon, self).__init__(
            jobs,
            job_trigger_state="parcel",
            job_function_name="task_upload_resources")


class VerifyDaemon(Daemon):
    def __init__(self, jobs):
        """Verify daemon"""
        super(VerifyDaemon, self).__init__(
            jobs,
            job_trigger_state="online",
            job_function_name="task_verify_resources")



def main(splash=True):
    import os
    import pkg_resources
    import sys

    from PyQt5.QtWidgets import QApplication

    app = QApplication(sys.argv)
    imdir = pkg_resources.resource_filename("dcoraid", "img")

    if splash:
        from PyQt5.QtWidgets import QSplashScreen
        from PyQt5.QtGui import QPixmap
        from PyQt5.QtCore import QEventLoop
        splash_path = os.path.join(imdir, "splash.png")
        splash_pix = QPixmap(splash_path)
        splash = QSplashScreen(splash_pix)
        splash.setMask(splash_pix.mask())
        splash.show()
        app.processEvents(QEventLoop.AllEvents, 300)

    import warnings
    from requests.packages.urllib3.exceptions import SubjectAltNameWarning
    # Ignore SubjectAltNameWarning for certificates in medical branding,
    # because they will show up in the dclab-compress-warnings log.
    warnings.filterwarnings("ignore", category=SubjectAltNameWarning)

    from PyQt5 import QtCore, QtGui
    from .gui import DCORAid

    # Set Application Icon
    icon_path = os.path.join(imdir, "icon.png")
    app.setWindowIcon(QtGui.QIcon(icon_path))

    # Use dots as decimal separators
    QtCore.QLocale.setDefault(QtCore.QLocale(QtCore.QLocale.C))

    window = DCORAid()

    if splash:
        splash.finish(window)

    sys.exit(app.exec_())


if __name__ == "__main__":
    main()



import functools
import time


def ttl_cache(seconds=5, maxsize=32, typed=False):
    """A time-to-live cache based on lru_cache"""
    def wrapper_cache(func):
        func = functools.lru_cache(maxsize=maxsize, typed=typed)(func)
        func.tinit = time.perf_counter()
        func.delta = 0

        @functools.wraps(func)
        def wrapped_func(*args, **kwargs):
            delta = (time.perf_counter() - func.tinit) // seconds
            if delta != func.delta:
                func.cache_clear()
                func.delta = delta
            return func(*args, **kwargs)
        return wrapped_func

    return wrapper_cache



import os
import pathlib
import pkg_resources
import signal
import sys
import traceback as tb

import appdirs
import dclab
import requests
import requests_toolbelt

from PyQt5 import uic, QtCore, QtGui, QtWidgets

from ..api import APIKeyError
from ..dbmodel import APIModel
from .._version import version as __version__

from .api import get_ckan_api
from .preferences import PreferencesDialog
from .tools import run_async
from .wizard import SetupWizard

# set Qt icon theme search path
QtGui.QIcon.setThemeSearchPaths([
    os.path.join(pkg_resources.resource_filename("dcoraid", "img"),
                 "icon-theme")])
QtGui.QIcon.setThemeName(".")


class StatusWidget(QtWidgets.QWidget):
    clicked = QtCore.pyqtSignal()

    def __init__(self, *args, **kwargs):
        super(StatusWidget, self).__init__(*args, **kwargs)
        self.layout = QtWidgets.QHBoxLayout(self)
        self.layout.setContentsMargins(0, 0, 0, 0)
        self.layout.setSpacing(0)

        self.flabel = QtWidgets.QLabel(self)
        self.layout.addWidget(self.flabel)

        self.toolButton_user = QtWidgets.QToolButton()
        self.toolButton_user.setText("Initialization...")
        self.toolButton_user.setToolButtonStyle(
            QtCore.Qt.ToolButtonTextBesideIcon)
        self.toolButton_user.setAutoRaise(True)
        self.layout.addWidget(self.toolButton_user)
        self.toolButton_user.clicked.connect(self.clicked)

    def get_favicon(self, server):
        dldir = pathlib.Path(appdirs.user_cache_dir(appname="DCOR-Aid",
                                                    appauthor="DCOR"))
        dldir.mkdir(exist_ok=True, parents=True)
        favname = dldir / (server.split("://")[1] + "_favicon.ico")
        if not favname.exists():
            try:
                r = requests.get(server + "/favicon.ico")
                if r.ok:
                    with favname.open("wb") as fd:
                        fd.write(r.content)
                else:
                    raise ValueError("No favicon!")
                favicon = QtGui.QIcon(str(favname))
            except BaseException:
                favicon = QtGui.QIcon()
        else:
            favicon = QtGui.QIcon(str(favname))
        return favicon

    def set_status(self, text, tooltip, icon, server):
        favicon = self.get_favicon(server)
        self.flabel.setPixmap(favicon.pixmap(16, 16))
        self.flabel.setToolTip(server)
        self.toolButton_user.setText(text)
        self.toolButton_user.setToolTip(tooltip)
        self.toolButton_user.setIcon(QtGui.QIcon.fromTheme(icon))


class DCORAid(QtWidgets.QMainWindow):
    plots_changed = QtCore.pyqtSignal()

    def __init__(self, *args, **kwargs):
        """Initialize DCOR_Manager

        If you pass the "--version" command line argument, the
        application will print the version after initialization
        and exit.
        """
        # Settings are stored in the .ini file format. Even though
        # `self.settings` may return integer/bool in the same session,
        # in the next session, it will reliably return strings. Lists
        # of strings (comma-separated) work nicely though.
        # Some promoted widgets need the below constants set in order
        # to access the settings upon initialization.
        QtCore.QCoreApplication.setOrganizationName("DCOR")
        QtCore.QCoreApplication.setOrganizationDomain("dcor.mpl.mpg.de")
        QtCore.QCoreApplication.setApplicationName("dcoraid")
        QtCore.QSettings.setDefaultFormat(QtCore.QSettings.IniFormat)
        #: DCOR-Aid settings
        self.settings = QtCore.QSettings()
        super(DCORAid, self).__init__(*args, **kwargs)
        path_ui = pkg_resources.resource_filename(
            "dcoraid.gui", "main.ui")
        uic.loadUi(path_ui, self)
        # if "--version" was specified, print the version and exit
        if "--version" in sys.argv:
            print(__version__)
            QtWidgets.QApplication.processEvents(QtCore.QEventLoop.AllEvents,
                                                 300)
            sys.exit(0)
        # GUI
        # Preferences dialog
        self.dlg_pref = PreferencesDialog()
        # Window title
        self.setWindowTitle("DCOR-Aid {}".format(__version__))
        # Disable native menubar (e.g. on Mac)
        self.menubar.setNativeMenuBar(False)
        # File menu
        self.actionPreferences.triggered.connect(self.dlg_pref.show_server)
        self.actionSetupWizard.triggered.connect(self.on_wizard)
        # Help menu
        self.actionSoftware.triggered.connect(self.on_action_software)
        self.actionAbout.triggered.connect(self.on_action_about)
        # Display login status
        self.status_widget = StatusWidget(self)
        self.tabWidget.setCornerWidget(self.status_widget)
        self.status_widget.clicked.connect(self.dlg_pref.on_show_server)
        self.refresh_login_status()
        # Call refresh_login status regularly
        self.timer = QtCore.QTimer()
        self.timer.timeout.connect(self.refresh_login_status)
        self.timer.start(300000)
        # Update private data tab
        self.refresh_private_data()
        # If a new dataset has been uploaded, refresh private data
        self.panel_upload.upload_finished.connect(self.refresh_private_data)

        if ((self.settings.value("user scenario", "") != "anonymous")
                and not self.settings.value("auth/api key", "")):
            QtWidgets.QApplication.processEvents(QtCore.QEventLoop.AllEvents,
                                                 300)
            # User has not done anything yet
            self.on_wizard()

        self.show()
        self.raise_()

    def close(self):
        self.timer.stop()
        self.panel_upload.widget_jobs.timer.stop()
        self.panel_upload.jobs.daemon_compress.join()
        self.panel_upload.jobs.daemon_upload.join()
        self.panel_upload.jobs.daemon_verify.join()
        self.panel_upload.jobs.daemon_compress.terminate()
        self.panel_upload.jobs.daemon_upload.terminate()
        self.panel_upload.jobs.daemon_verify.terminate()
        QtWidgets.QApplication.processEvents(QtCore.QEventLoop.AllEvents,
                                             3000)
        super(DCORAid, self).close()

    def on_action_about(self):
        dcor = "https://dcor.mpl.mpg.de"
        gh = "DCOR-dev/DCOR-Aid"
        rtd = "dc.readthedocs.io"
        about_text = "This is the client for the <a href='{}'>".format(dcor) \
            + "Deformability Cytometry Open Repository (DCOR)</a>.<br><br>" \
            + "Author: Paul Müller<br>" \
            + "GitHub: " \
            + "<a href='https://github.com/{gh}'>{gh}</a><br>".format(gh=gh) \
            + "Documentation: " \
            + "<a href='https://{rtd}'>{rtd}</a><br>".format(rtd=rtd)
        QtWidgets.QMessageBox.about(self,
                                    "DCOR-Aid {}".format(__version__),
                                    about_text)

    def on_action_software(self):
        libs = [appdirs,
                dclab,
                requests,
                requests_toolbelt,
                ]
        sw_text = "DCOR-Aid {}\n\n".format(__version__)
        sw_text += "Python {}\n\n".format(sys.version)
        sw_text += "Modules:\n"
        for lib in libs:
            sw_text += "- {} {}\n".format(lib.__name__, lib.__version__)
        sw_text += "- PyQt5 {}\n".format(QtCore.QT_VERSION_STR)
        sw_text += "\n Breeze icon theme by the KDE Community (LGPL)."
        sw_text += "\n Font-Awesome icons by Fort Awesome (CC BY 4.0)."
        if hasattr(sys, 'frozen'):
            sw_text += "\nThis executable has been created using PyInstaller."
        QtWidgets.QMessageBox.information(self,
                                          "Software",
                                          sw_text)

    @QtCore.pyqtSlot()
    def on_wizard(self):
        self.wizard = SetupWizard(self)
        self.wizard.exec_()

    @run_async
    @QtCore.pyqtSlot()
    def refresh_login_status(self):
        api = get_ckan_api()
        if not api.is_available():
            text = "No connection"
            tip = "Can you access {} via a browser?".format(api.server)
            icon = "hourglass"
        else:
            if not api.api_key:
                text = "Anonymous"
                tip = "Click here to enter your API key."
                icon = "user"
            else:
                try:
                    user_data = api.get_user_dict()
                except APIKeyError:
                    text = "Login failed"
                    tip = "Click here to update your API key."
                    icon = "user-times"
                else:
                    fullname = user_data["fullname"]
                    name = user_data["name"]
                    if not fullname:
                        fullname = name
                    text = "{}".format(fullname)
                    tip = "user '{}'".format(name)
                    icon = "user-lock"
        self.status_widget.set_status(text=text,
                                      tooltip=tip,
                                      icon=icon,
                                      server=api.server)

    @run_async
    @QtCore.pyqtSlot()
    def refresh_private_data(self):
        self.tab_user.setCursor(QtCore.Qt.WaitCursor)
        # TODO:
        # - what happens if the user changes the server? Ask to restart?
        api = get_ckan_api()
        try:
            am = APIModel(api=api)
            if am.api.is_available():
                db_extract = am.get_user_datasets()
        except BaseException:
            pass
        else:
            self.user_filter_chain.set_db_extract(db_extract)
        self.tab_user.setCursor(QtCore.Qt.ArrowCursor)


def excepthook(etype, value, trace):
    """
    Handler for all unhandled exceptions.

    :param `etype`: the exception type (`SyntaxError`,
        `ZeroDivisionError`, etc...);
    :type `etype`: `Exception`
    :param string `value`: the exception error message;
    :param string `trace`: the traceback header, if any (otherwise, it
        prints the standard Python header: ``Traceback (most recent
        call last)``.
    """
    vinfo = "Unhandled exception in DCOR-Aid version {}:\n".format(
        __version__)
    tmp = tb.format_exception(etype, value, trace)
    exception = "".join([vinfo] + tmp)

    errorbox = QtWidgets.QMessageBox()
    errorbox.setIcon(QtWidgets.QMessageBox.Critical)
    errorbox.addButton(QtWidgets.QPushButton('Close'),
                       QtWidgets.QMessageBox.YesRole)
    errorbox.addButton(QtWidgets.QPushButton(
        'Copy text && Close'), QtWidgets.QMessageBox.NoRole)
    errorbox.setText(exception)
    ret = errorbox.exec_()
    if ret == 1:
        cb = QtWidgets.QApplication.clipboard()
        cb.clear(mode=cb.Clipboard)
        cb.setText(exception)


# Make Ctr+C close the app
signal.signal(signal.SIGINT, signal.SIG_DFL)
# Display exception hook in separate dialog instead of crashing
sys.excepthook = excepthook



import pkg_resources
import webbrowser

from PyQt5 import uic, QtCore, QtWidgets


class TableCellActions(QtWidgets.QWidget):
    abort_job = QtCore.pyqtSignal(str)
    delete_job = QtCore.pyqtSignal(str)

    def __init__(self, job, *args, **kwargs):
        """Actions in a table cell

        Used for the "Running Uploads" table in the "Uploads" tab.
        """
        super(TableCellActions, self).__init__(*args, **kwargs)
        path_ui = pkg_resources.resource_filename(
            "dcoraid.gui.upload", "widget_tablecell_actions.ui")
        uic.loadUi(path_ui, self)
        self.job = job
        # signals and slots
        self.tb_abort.clicked.connect(self.on_abort)
        self.tb_delete.clicked.connect(self.on_delete)
        self.tb_error.clicked.connect(self.on_error)
        self.tb_retry.clicked.connect(self.on_retry)
        self.tb_view.clicked.connect(self.on_view)

    @QtCore.pyqtSlot()
    def on_abort(self):
        self.abort_job.emit(self.job.dataset_id)

    @QtCore.pyqtSlot()
    def on_delete(self):
        self.delete_job.emit(self.job.dataset_id)

    @QtCore.pyqtSlot()
    def on_error(self):
        msg = QtWidgets.QMessageBox()
        msg.setIcon(QtWidgets.QMessageBox.Critical)
        msg.setText("There was an error during data transfer. If this happens "
                    + "often or with a particular type of dataset, please "
                    + "<a href='https://github.com/DCOR-dev/DCOR-Aid/issues'>"
                    + "create an issue online</a>.")
        msg.setWindowTitle("Job {} error".format(self.job.dataset_id[:5]))
        msg.setDetailedText(self.job.traceback)
        msg.exec_()

    @QtCore.pyqtSlot()
    def on_retry(self):
        self.job.retry_upload()

    @QtCore.pyqtSlot()
    def on_view(self):
        url = self.job.get_dataset_url()
        webbrowser.open(url)

    def refresh_visibility(self, job):
        """Show or hide the different toolbuttons depending on the job state"""
        self.job = job
        state = job.state

        if state in ["online", "verify", "finalize", "done"]:
            self.tb_view.show()
        else:
            self.tb_view.hide()

        if state in ["abort", "error"]:
            self.tb_retry.show()
        else:
            self.tb_retry.hide()

        if state == "error":
            self.tb_error.show()
        else:
            self.tb_error.hide()

        if state == "transfer":
            self.tb_abort.show()
        else:
            self.tb_abort.hide()

        if state not in ["compress", "transfer"]:
            self.tb_delete.show()
        else:
            self.tb_delete.hide()



from functools import lru_cache
import hashlib
import pathlib
import shutil
import time
import traceback as tb
import warnings

import appdirs
from dclab.rtdc_dataset.check import IntegrityChecker
from dclab.cli import compress

from ..api import APIError

from . import dataset


#: Valid job states (in more or less chronological order)
JOB_STATES = [
    "init",  # inital
    "compress",  # compression (only for DC data)
    "parcel",  # ready for upload
    "transfer",  # upload in progress
    "online",  # dataset has been transferred
    "verify",  # perform SHA256 sum verification
    "finalize",  # CKAN dataset is being activated
    "done",  # job finished
    "abort",  # user aborted
    "error",  # error occured
]


class UploadJob(object):
    def __init__(self, dataset_dict, paths, api,
                 resource_names=None, supplements=None):
        """Wrapper for resource uploads

        This job is meant to be run from a separate thread.
        The function `task_compress_resources` ensures that all
        resources are compressed. The function `task_upload_resources`
        performs the actual upload. During upload, the progress
        is monitored and can be read from other threads
        using the e.g. `get_status` function.
        """
        self.dataset_dict = dataset_dict
        self.dataset_id = dataset_dict["id"]
        self.api = api.copy()  # create a copy of the API
        self.paths = paths
        if resource_names is None:
            resource_names = [pathlib.Path(pp).name for pp in paths]
        #: Important supplementary resource schema meta data that will
        #: be formatted to composite {"sp:section:key" = value} an appended
        #: to the resource metadata.
        self.supplements = supplements
        self.resource_names = resource_names
        self.paths_uploaded = []
        self.set_state("init")
        self.traceback = None
        self.file_sizes = [pathlib.Path(ff).stat().st_size for ff in paths]
        self.file_bytes_uploaded = [0] * len(paths)
        self.index = 0
        self.start_time = None
        self.end_time = None
        self._last_time = 0
        self._last_bytes = 0
        self._last_rate = 0
        # caching
        dcoraid_cache = pathlib.Path(appdirs.user_cache_dir()) / "dcoraid"
        self.cache_dir = dcoraid_cache / "compress-{}".format(self.dataset_id)

    def cleanup(self):
        """cleanup temporary files in the user's cache directory"""
        shutil.rmtree(self.cache_dir, ignore_errors=True)

    def get_composite_supplements(self, resource_index):
        """Return the composite supplements "sp:section:key" for a resource"""
        sp = {}
        if self.supplements:
            sp_dict = self.supplements[resource_index]
            for sec in sp_dict:
                for key in sp_dict[sec]:
                    sp["sp:{}:{}".format(sec, key)] = sp_dict[sec][key]
        return sp

    def get_dataset_url(self):
        """Return a link to the dataset on DCOR"""
        # The API prepends missing "https://"
        return "{}/dataset/{}".format(self.api.server, self.dataset_id)

    def get_progress_string(self):
        """Return a nice string representation of the progress"""
        status = self.get_status()
        state = status["state"]
        plural = "s" if status["files total"] > 1 else ""

        if state in ["init", "compress", "parcel"]:
            progress = "0% ({} file{})".format(status["files total"], plural)
        elif state == "transfer":
            progress = "{:.0f}% (file {}/{})".format(
                status["bytes uploaded"]/status["bytes total"]*100,
                status["files uploaded"]+1,
                status["files total"])
        elif state in ["finalize", "online", "verify", "done"]:
            progress = "100% ({} file{})".format(status["files total"],
                                                 plural)
        elif state in ["abort", "error"]:
            progress = "-- ({}/{} file{})".format(status["files uploaded"],
                                                  status["files total"],
                                                  plural)
        return progress

    def get_rate_string(self):
        """Return a nice string representing upload rate"""
        status = self.get_status()
        state = status["state"]
        rate = status["rate"]
        if state in ["init", "compress", "parcel"]:
            rate_label = "-- kB/s"
        else:
            if rate > 1e6:
                rate_label = "{:.1f} MB/s".format(rate/1e6)
            else:
                rate_label = "{:.0f} kB/s".format(rate/1e3)
            if state != "transfer":
                rate_label = "⌀ " + rate_label
        return rate_label

    def get_rate(self, resolution=1):
        """Return the current resource upload rate

        Parameters
        ----------
        resolution: float
            Time interval in which to perform the average.

        Returns
        -------
        upload_rate: float
            Mean upload rate in bytes per second
        """
        cur_time = time.perf_counter()
        cur_bytes = sum(self.file_bytes_uploaded)

        delta_time = cur_time - self._last_time
        delta_bytes = cur_bytes - self._last_bytes

        if self.start_time is None:
            # not started yet
            rate = 0
            self._last_bytes = 0
        elif self.end_time is None:
            # not finished yet
            if delta_time > resolution:
                rate = delta_bytes / delta_time
                self._last_time = cur_time
                self._last_bytes = cur_bytes
            else:
                rate = self._last_rate
        else:
            # finished
            tdelt = (self.end_time - self.start_time)
            rate = sum(self.file_bytes_uploaded) / tdelt
        self._last_rate = rate
        return rate

    def get_status(self):
        """Get the status of the current job

        Returns
        -------
        status: dict
            Dictionary with various interesting parameters
        """
        data = {
            "state": self.state,
            "files total": len(self.paths),
            "files uploaded": len(self.paths_uploaded),
            "bytes total": sum(self.file_sizes),
            "bytes uploaded": sum(self.file_bytes_uploaded),
            "rate": self.get_rate(),
        }
        return data

    def monitor_callback(self, monitor):
        """Upload progress monitor callback function

        This method updates `self.file_sizes` and
        `self.file_bytes_uploaded` for the currently
        uploading resource.
        """
        self.file_sizes[self.index] = monitor.len
        self.file_bytes_uploaded[self.index] = monitor.bytes_read

    def retry_upload(self):
        """Retry uploading resources when an error occured"""
        if self.state in ["abort", "error"]:
            self.set_state("parcel")
        else:
            raise ValueError("Can only retry upload in error state!")

    def set_state(self, state):
        """Set the current job state

        The state is checked against :const:`JOB_STATES`
        """
        if state not in JOB_STATES:
            raise ValueError("Unknown state: '{}'".format(state))
        self.state = state

    def task_compress_resources(self):
        """Compress resources if they are not fully compressed

        Data are stored in the user's cache directory and
        deleted after upload is complete.
        """
        self.set_state("compress")
        for ii, path in enumerate(list(self.paths)):
            if path.suffix in [".rtdc", ".dc"]:  # do we have an .rtdc file?
                # check for compression
                with IntegrityChecker(path) as ic:
                    cdata = ic.check_compression()[0].data
                if cdata["uncompressed"]:  # (partially) not compressed?
                    res_dir = self.cache_dir / "{}".format(ii)
                    res_dir.mkdir(exist_ok=True, parents=True)
                    path_out = res_dir / path.name
                    compress(path_out=path_out, path_in=path)
                    self.paths[ii] = path_out
        self.set_state("parcel")

    def task_upload_resources(self):
        """Start the upload

        The progress of the upload is monitored and written
        to attributes. The current status can be retrieved
        via :func:`UploadJob.get_status`.
        """
        if self.state == "parcel":
            try:
                self.set_state("transfer")
                self.start_time = time.perf_counter()
                # Do the things to do and watch self.state while doing so
                for ii, path in enumerate(self.paths):
                    self.index = ii
                    resource_name = self.resource_names[ii]
                    resource_supplement = self.get_composite_supplements(ii)
                    exists = dataset.resource_exists(
                        dataset_id=self.dataset_id,
                        resource_name=resource_name,
                        api=self.api)
                    if exists:
                        # We are currently retrying an upload. If the
                        # resource exists, we have already uploaded it.
                        self.file_bytes_uploaded[ii] = self.file_sizes[ii]
                        continue
                    if not exists:
                        try:
                            dataset.add_resource(
                                dataset_id=self.dataset_id,
                                path=path,
                                resource_name=resource_name,
                                resource_dict=resource_supplement,
                                api=self.api,
                                monitor_callback=self.monitor_callback)
                            self.paths_uploaded.append(path)
                        except APIError:
                            # Workaround for large datasets (no response)
                            # just check whether the resource is there
                            exists = dataset.resource_exists(
                                dataset_id=self.dataset_id,
                                resource_name=resource_name,
                                api=self.api)
                            if not exists:
                                raise
                        except SystemExit:
                            # This thread has just been killed
                            self.start_time = None
                            self.file_bytes_uploaded[ii] = 0
                            self.set_state("abort")
                            return
                self.end_time = time.perf_counter()
                self.set_state("online")
            except BaseException:
                self.set_state("error")
                self.traceback = tb.format_exc()
        else:
            warnings.warn("Starting an upload only possible when state is "
                          + "'parcel', but current state is "
                          + "'{}'!".format(self.state))

    def task_verify_resources(self):
        """Perform SHA256 verification"""
        if self.state == "online":
            # First check whether all SHA256 sums are already available online
            sha256dict = dataset.resource_sha256_sums(
                dataset_id=self.dataset_id,
                api=self.api)
            if sum([sha256dict[name] is None for name in sha256dict]) != 0:
                # only start verification if all SHA256 sums are available
                pass
            else:
                self.set_state("verify")
                for ii, path in enumerate(self.paths):
                    resource_name = self.resource_names[ii]
                    # compute SHA256 sum
                    sha = sha256sum(path)
                    if sha != sha256dict[resource_name]:
                        self.set_state("error")
                        self.traceback = "SHA256 sum failed for resource " \
                            + "'{}' ({} vs. {})!".format(
                                resource_name, sha, sha256dict[resource_name])
                        break
                else:
                    self.cleanup()
                    # finalize dataset
                    self.set_state("finalize")
                    # draft -> active
                    dataset.activate_dataset(
                        dataset_id=self.dataset_id,
                        api=self.api)
                    self.set_state("done")
        else:
            warnings.warn("Resource verification is only possible when state "
                          + "is 'online', but current state is "
                          + "'{}'!".format(self.state))


@lru_cache(maxsize=2000)
def sha256sum(path):
    """Compute the SHA256 sum of a file on disk"""
    block_size = 2**20
    path = pathlib.Path(path)
    file_hash = hashlib.sha256()
    with path.open("rb") as fd:
        while True:
            data = fd.read(block_size)
            if not data:
                break
            file_hash.update(data)
    return file_hash.hexdigest()



import pathlib
import time

from dcoraid.upload import dataset
from dcoraid.api import APIConflictError

import common


dpath = pathlib.Path(__file__).parent / "data" / "calibration_beads_47.rtdc"


def make_dataset_dict(hint=""):
    hint += " "
    dataset_dict = {
        "title": "A test dataset {}{}".format(hint, time.time()),
        "private": True,
        "license_id": "CC0-1.0",
        "owner_org": common.CIRCLE,
        "authors": common.USER_NAME,
    }
    return dataset_dict


def test_dataset_create_same_resource():
    """There should be an error when a resource is added twice"""
    api = common.get_api()
    # create some metadata
    dataset_dict = make_dataset_dict(hint="create-with-same-resource")
    # post dataset creation request
    data = dataset.create_dataset(dataset_dict=dataset_dict,
                                  api=api)
    dataset.add_resource(dataset_id=data["id"],
                         path=dpath,
                         api=api
                         )
    try:
        dataset.add_resource(dataset_id=data["id"],
                             path=dpath,
                             api=api
                             )
    except APIConflictError:
        pass
    else:
        assert False, "Should not be able to upload same resource twice"


def test_dataset_creation():
    """Just test whether we can create (and remove) a draft dataset"""
    api = common.get_api()
    # create some metadata
    dataset_dict = make_dataset_dict(hint="basic_test")
    # post dataset creation request
    data = dataset.create_dataset(dataset_dict=dataset_dict,
                                  api=api,
                                  )
    # simple test
    assert "authors" in data
    assert data["authors"] == common.USER_NAME
    assert data["state"] == "draft"
    # remove draft dataset
    dataset.remove_draft(dataset_id=data["id"],
                         api=api,
                         )
    # make sure it is gone
    try:
        api.get("package_show", id=data["id"])
    except BaseException:
        pass
    else:
        assert False


if __name__ == "__main__":
    # Run all tests
    loc = locals()
    for key in list(loc.keys()):
        if key.startswith("test_") and hasattr(loc[key], "__call__"):
            loc[key]()



import copy
import pathlib

from requests_toolbelt import MultipartEncoder, MultipartEncoderMonitor


def activate_dataset(dataset_id, api):
    """Change the state of a dataset to "active"

    In the DCOR workflow, datasets are created as drafts and
    resources are added to the drafts. After that, the
    datasets are activated (and no resources can be added
    anymore).

    Parameters
    ----------
    dataset_id: str
        CKAN ID of the dataset
    api: dcoraid.api.CKANAPI
        API instance with server, api_key (and optional certificate)
    """
    revise_dict = {
        "match": {"id": dataset_id},
        "update": {"state": "active"}}
    res = api.post("package_revise", revise_dict)
    if res["package"]["state"] != "active":
        # TODO: Make this an error message in ckanext-dcor_schemas
        raise ValueError("Setting the state of the dataset "
                         + "'{}' to 'active' did not work! ".format(dataset_id)
                         + "Does the dataset contain resources? If not, then "
                         + "you are experienceing a ckanext-dcor_schemas "
                         + "restriction.")


def add_resource(dataset_id, path, api, resource_name=None,
                 resource_dict=None, monitor_callback=None):
    """Add a resource to a dataset

    Parameters
    ----------
    dataset_id: str
        CKAN ID of the dataset to which the resource is added
    path: str or pathlib.Path
        Path to the resource
    api: dcoraid.api.CKANAPI
        API instance with server, api_key (and optional certificate)
    resource_name: str
        Alternate resource name, if not set `path.name` is used
    resource_dict: dict
        Dictionary of resource meta data (used for supplementary
        resource schemas "sp:section:key")
    monitor_callback: None or callable
        This callable is used to monitor the upload progress. It must
        accept one argument: a
        `requests_toolbelt.MultipartEncoderMonitor` object.

    See Also
    --------
    dcoraid.upload.joblist.UploadJob
        An implementation of an upload job that monitors progress.
    """
    if resource_dict is None:
        resource_dict = {}
    path = pathlib.Path(path)
    if resource_name is None:
        resource_name = path.name
    e = MultipartEncoder(fields={
        'package_id': dataset_id,
        'name': resource_name,
        'upload': (resource_name, path.open('rb'))})
    m = MultipartEncoderMonitor(e, monitor_callback)
    # perform upload
    data = api.post("resource_create",
                    data=m,
                    dump_json=False,
                    headers={"Content-Type": m.content_type})
    if resource_dict:
        # add resource_dict
        revise_dict = {
            "match": {"id": dataset_id},
            "update__resources__{}".format(data["id"]): resource_dict}
        api.post("package_revise", revise_dict)


def create_dataset(dataset_dict, api, resources=[],
                   create_circle=False, activate=False):
    """Create a draft dataset

    Parameters
    ----------
    dataset_dict: dict
        CKAN dataset dictionary
    api: dcoraid.api.CKANAPI
        API instance with server, api_key (and optional certificate)
    resources: list of str or pathlib.Path
        Paths to dataset resources
    create_circle: bool
        Create the circle if it does not already exist
    activate: bool
        If True, then the dataset state is changed to "active"
        after uploading of the resources is complete. For DCOR,
        this implies that no other resources can be added to the
        dataset.
    """
    if create_circle:
        circles = [c["name"] for c in api.get("organization_list_for_user")]
        if dataset_dict["owner_org"] not in circles:
            # Create the circle before creating the dataset
            api.post("organization_create",
                     data={"name": dataset_dict["owner_org"]})
    dataset_dict = copy.deepcopy(dataset_dict)
    dataset_dict["state"] = "draft"
    data = api.post("package_create", dataset_dict)
    if resources:
        for res in resources:
            add_resource(dataset_id=data["id"],
                         path=res,
                         api=api)
    if activate:
        activate_dataset(dataset_id=data["id"], api=api)
        data["state"] = "active"
    return data


def remove_draft(dataset_id, api):
    """Remove a draft dataset

    Use this for deleting datasets that you created but which were
    not yet activated. The dataset is deleted and purged.

    Parameters
    ----------
    dataset_id: str
        CKAN ID of the dataset to which the resource is added
    api: dcoraid.api.CKANAPI
        API instance with server, api_key (and optional certificate)
    """
    api.post("package_delete", {"id": dataset_id})
    api.post("dataset_purge", {"id": dataset_id})


def remove_all_drafts(api):
    """Remove all draft datasets

    Find and delete all draft datasets for a user. The user
    ID is inferred from the API key.

    Parameters
    ----------
    api: dcoraid.api.CKANAPI
        API instance with server, api_key (and optional certificate)
    """
    user_dict = api.get_user_dict()
    data = api.get(
        "package_search",
        q="*:*",
        include_drafts=True,
        include_private=True,
        fq="creator_user_id:{} AND state:draft".format(user_dict["id"]),
        rows=1000)
    for dd in data["results"]:
        assert dd["state"] == "draft"
        remove_draft(dd["id"], api=api)
    return data["results"]


def resource_exists(dataset_id, resource_name, api):
    """Check whether a resource exists in a dataset"""
    pkg_dict = api.get("package_show", id=dataset_id)
    for resource in pkg_dict["resources"]:
        if resource["name"] == resource_name:
            return True
    else:
        return False


def resource_sha256_sums(dataset_id, api):
    """Return a dictionary of resources with the SHA256 sums as values"""
    pkg_dict = api.get("package_show", id=dataset_id)
    sha256dict = {}
    for resource in pkg_dict["resources"]:
        sha256dict[resource["name"]] = resource.get("sha256", None)
    return sha256dict



import time
from dcoraid.dbmodel import util


def test_cache():
    count = 0

    @util.ttl_cache(seconds=.1)
    def test(arg1):
        nonlocal count
        count += 1
        return count

    assert test(1) == 1, "called the first time we invoke it"
    assert test(1) == 1, "not called because it is already cached"

    time.sleep(.01)
    assert test(1) == 1, "not called because it is already cached"

    # Let's now wait for the cache to expire
    time.sleep(.1)
    assert test(1) == 2, "called because the cache expired"


if __name__ == "__main__":
    # Run all tests
    loc = locals()
    for key in list(loc.keys()):
        if key.startswith("test_") and hasattr(loc[key], "__call__"):
            loc[key]()



import time

from dcoraid.gui.main import DCORAid
from dcoraid.gui.upload.dlg_upload import UploadDialog

import pytest
from PyQt5 import QtCore, QtWidgets
from PyQt5.QtWidgets import QMessageBox

import common


@pytest.fixture(autouse=True)
def run_around_tests():
    # Code that will run before your test, for example:
    pass
    # Run test
    yield
    # Make sure that all daemons are gone
    time.sleep(2)
    QtWidgets.QApplication.processEvents(QtCore.QEventLoop.AllEvents,
                                         3000)
    time.sleep(2)


def test_upload_simple(qtbot, monkeypatch):
    """Upload a test dataset"""
    mw = DCORAid()
    QtWidgets.QApplication.processEvents(QtCore.QEventLoop.AllEvents, 300)

    dlg = UploadDialog(mw.panel_upload)
    dlg.finished.connect(mw.panel_upload.on_run_upload)
    # Fill data for testing
    dlg._autofill_for_testing()
    # Avoid message boxes
    monkeypatch.setattr(QMessageBox, "question", lambda *args: QMessageBox.Yes)
    # Commence upload
    dlg.on_proceed()
    assert dlg.dataset_id is not None
    for ii in range(200):  # give it 20secs to upload
        state = mw.panel_upload.jobs[0].get_status()["state"]
        if state == "done":
            break
        time.sleep(.1)
    else:
        raise ValueError("Job did not complete, state: '{}'".format(state))
    mw.close()


def test_upload_private(qtbot, monkeypatch):
    """Upload a private test dataset"""
    mw = DCORAid()
    QtWidgets.QApplication.processEvents(QtCore.QEventLoop.AllEvents, 300)

    dlg = UploadDialog(mw.panel_upload)
    dlg.finished.connect(mw.panel_upload.on_run_upload)
    # Fill data for testing
    dlg._autofill_for_testing()
    # set visibility to private
    dlg.comboBox_vis.setCurrentIndex(dlg.comboBox_vis.findData("private"))
    # Avoid message boxes
    monkeypatch.setattr(QMessageBox, "question", lambda *args: QMessageBox.Yes)
    # Commence upload
    dlg.on_proceed()
    dataset_id = dlg.dataset_id
    assert dataset_id is not None
    for ii in range(200):  # give it 20secs to upload
        state = mw.panel_upload.jobs[0].get_status()["state"]
        if state == "done":
            break
        time.sleep(.1)
    else:
        raise ValueError("Job did not complete, state: '{}'".format(state))
    mw.close()
    # make sure the dataset is private
    api = common.get_api()
    dataset_dict = api.get(api_call="package_show", id=dataset_id)
    assert dataset_dict["private"]
    assert isinstance(dataset_dict["private"], bool)



from functools import lru_cache
import pathlib
import pkg_resources
import traceback as tb

from PyQt5 import uic, QtCore, QtGui, QtWidgets

from ...upload import create_dataset

from ..api import get_ckan_api

from .resources_model import ResourcesModel
from .resource_schema_preset import PersistentResourceSchemaPresets


class UploadDialog(QtWidgets.QDialog):
    finished = QtCore.pyqtSignal(object)
    instance_counter = 1

    def __init__(self, parent=None):
        """Create a new window for setting up a file upload
        """
        super(UploadDialog, self).__init__(parent)
        path_ui = pkg_resources.resource_filename(
            "dcoraid.gui.upload", "dlg_upload.ui")
        uic.loadUi(path_ui, self)

        # Dialog box buttons
        self.pushButton_proceed = self.buttonBox.button(
            QtWidgets.QDialogButtonBox.Apply)
        self.pushButton_proceed.setText("Proceed with upload / Enqueue job")
        self.pushButton_cancel = self.buttonBox.button(
            QtWidgets.QDialogButtonBox.Cancel)

        # Keep identifier
        self.identifier = self.instance_counter
        UploadDialog.instance_counter += 1

        #: This will be overridden with the actual value assigned
        #: by DCOR. It is important for later uploading the resources.
        self.dataset_id = None

        self.setWindowTitle("DCOR Upload {}".format(self.identifier))

        # Initialize api
        self.api = get_ckan_api()

        # Set license choices
        licenses = self.api.get_license_list()
        for lic in licenses:
            if lic["domain_data"]:  # just some identifier to exclude "none"
                self.comboBox_license.addItem(
                    "{} ({})".format(lic["title"], lic["id"]), lic["id"])

        # Set supplementary resource schema
        rss = self.api.get_supplementary_resource_schema()
        self.widget_schema.populate_schema(rss)

        # Set circle choices
        circles = self.get_user_circle_dicts()
        for ci in circles:
            self.comboBox_circles.addItem(
                ci["title"] if ci["title"] else ci["name"], ci["name"])

        # Set visibility choices
        settings = QtCore.QSettings()
        if settings.value("user scenario", "") == "medical":
            # only allow private datasets
            self.comboBox_vis.addItem("Private", "private")
        else:
            self.comboBox_vis.addItem("Public", "public")
            self.comboBox_vis.addItem("Private", "private")

        # Shortcut for testing
        self.shortcut = QtWidgets.QShortcut(
            QtGui.QKeySequence("Ctrl+Alt+Shift+E"), self)
        self.shortcut.activated.connect(self._autofill_for_testing)

        # Setup resources view
        self.rvmodel = ResourcesModel()
        self.listView_resources.setModel(self.rvmodel)
        self.selModel = self.listView_resources.selectionModel()

        # Effectively hide resource schema options initially
        self.on_selection_changed()

        # Presets for user convenience
        self.presets = PersistentResourceSchemaPresets()
        self.comboBox_preset.addItems(sorted(self.presets.keys()))

        # Signals and slots
        # general buttons
        self.toolButton_add.clicked.connect(self.on_add_resources)
        self.toolButton_rem.clicked.connect(self.on_rem_resources)
        self.pushButton_proceed.clicked.connect(self.on_proceed)
        # resource-related signals
        self.lineEdit_res_filename.textChanged.connect(
            self.on_update_resources_model)
        self.widget_schema.schema_changed.connect(
            self.on_update_resources_model)
        self.selModel.selectionChanged.connect(self.on_selection_changed)
        # do not allow to proceed without a title
        self.lineEdit_authors.textChanged.connect(self.on_authors_edited)
        self.on_authors_edited("")  # initial state
        # if the user changes the preset combobox text, offer him
        # to save it as a preset
        self.comboBox_preset.editTextChanged.connect(
            self.on_preset_text_edited)
        # also offer the user to save the preset if the schema changed
        self.widget_schema.schema_changed.connect(
            self.on_preset_text_edited)
        self.on_preset_text_edited()  # empty string means no preset
        # store/load a preset
        self.toolButton_preset_load.clicked.connect(self.on_preset_load)
        self.toolButton_preset_store.clicked.connect(self.on_preset_store)

    def _autofill_for_testing(self, **kwargs):
        self.lineEdit_title.setText(kwargs.get("title", "Dataset Title"))
        self.lineEdit_authors.setText(kwargs.get("authors", "John Doe"))
        self.lineEdit_doi.setText(kwargs.get("doi", ""))
        self.lineEdit_references.setText(kwargs.get("references", ""))
        self.plainTextEdit_notes.setPlainText(
            kwargs.get("notes", "A description"))
        self.lineEdit_tags.setText(kwargs.get("tags", "HL60, GFP"))
        licen = "CC-BY-SA-4.0"
        self.comboBox_license.setCurrentIndex(
            self.comboBox_license.findData(licen))
        self.comboBox_circles.setCurrentIndex(0)
        self.comboBox_vis.setCurrentIndex(0)
        relpath = "../../../tests/data/calibration_beads_47.rtdc"
        path = pathlib.Path(__file__).resolve().parent / relpath
        if path.exists():
            self.on_add_resources([str(path.resolve())])

    @lru_cache(maxsize=1)
    def get_user_circle_dicts(self):
        circles = self.api.get("organization_list_for_user",
                               permission="create_dataset")
        if not circles:
            # Ask the user whether he would like to create a circle.
            ud = self.api.get_user_dict()
            name = ud["fullname"] if ud["fullname"] else ud["name"]
            text, okPressed = QtWidgets.QInputDialog.getText(
                self,
                "Circle required",
                "You do not have access to any existing Circles. To upload\n"
                + "datasets, you need to be either Editor or Admin in\n"
                + "a Circle. You may create a Circle now or cancel and ask\n"
                + "a colleague to add you to a Circle (Your user name is "
                + "'{}').".format(ud["name"])
                + "\n\nTo proceed with Circle creation, please choose a name:",
                QtWidgets.QLineEdit.Normal,
                "{}'s Circle".format(name))
            if okPressed and text != '':
                cname = "user-circle-{}".format(ud["name"])
                cdict = self.api.post("organization_create",
                                      data={"name": cname,
                                            "title": text.strip(),
                                            })
                circles.append(cdict)
            else:
                self.deleteLater()
                self.close()
        return circles

    def assemble_metadata(self):
        """Get all the metadata from the form"""
        # Dataset
        tags = []
        for tt in self.lineEdit_tags.text().replace(" ", "").split(","):
            if tt:
                tags.append({"name": tt})

        dataset_dict = {
            "title": self.lineEdit_title.text(),
            "authors": self.lineEdit_authors.text(),
            "doi": self.lineEdit_doi.text(),
            "references": self.lineEdit_references.text(),
            "tags": tags,
            "notes": self.plainTextEdit_notes.toPlainText(),
            "license_id": self.comboBox_license.currentData(),
            "private": self.comboBox_vis.currentData() == "private",
            "owner_org": self.comboBox_circles.currentData(),
        }
        return dataset_dict

    @QtCore.pyqtSlot(str)
    def on_authors_edited(self, newtext):
        """Enable proceed button if authors text is set"""
        if newtext:
            enabled = True
            tooltip = "Proceed with upload"
        else:
            enabled = False
            tooltip = "Set 'Authors' to proceed with upload"
        self.pushButton_proceed.setEnabled(enabled)
        self.pushButton_proceed.setToolTip(tooltip)

    @QtCore.pyqtSlot()
    def on_add_resources(self, files=None):
        """Ask the user to specify files to add"""
        if files is None:
            suffixes = self.api.get_supported_resource_suffixes()
            files, _ = QtWidgets.QFileDialog.getOpenFileNames(
                self, "Resources to upload", ".",
                "Supported file types ({})".format(
                    " ".join(["*{}".format(s) for s in suffixes])))
        self.rvmodel.add_resources(files)

        if not self.listView_resources.selectedIndexes():
            # Select the first item
            ix = self.rvmodel.index(0, 0)
            sm = self.listView_resources.selectionModel()
            sm.select(ix, QtCore.QItemSelectionModel.Select)

    @QtCore.pyqtSlot()
    def on_preset_load(self):
        """Load the preset with the current name"""
        curtext = self.comboBox_preset.currentText()

        # Load from presets (must come before UI logic)
        self.widget_schema.set_schema(self.presets[curtext])

        # UI logic
        self.toolButton_preset_store.setEnabled(False)
        self.toolButton_preset_load.setEnabled(False)

    @QtCore.pyqtSlot()
    def on_preset_store(self):
        """Store the current preset"""
        curtext = self.comboBox_preset.currentText()

        # Store preset
        preset = self.widget_schema.get_current_schema()
        self.presets[curtext] = preset
        self.on_update_resources_model()

        # UI logic
        items = []
        for ii in range(self.comboBox_preset.count()):
            items.append(self.comboBox_preset.itemText(ii))
        items = sorted(set(items + [curtext]))
        self.comboBox_preset.blockSignals(True)
        self.comboBox_preset.clear()
        self.comboBox_preset.addItems(items)
        self.comboBox_preset.setCurrentIndex(items.index(curtext))
        self.comboBox_preset.blockSignals(False)
        self.toolButton_preset_store.setEnabled(False)
        self.toolButton_preset_load.setEnabled(False)

    @QtCore.pyqtSlot()
    def on_preset_text_edited(self):
        """The preset combobox text was edited by the user

        Give or take a away from the user the option to load or save
        a preset.
        """
        # UI logic only
        curtext = self.comboBox_preset.currentText()
        if curtext:
            self.toolButton_preset_store.setEnabled(True)
        else:
            self.toolButton_preset_store.setEnabled(False)
        if curtext in self.presets:
            self.toolButton_preset_load.setEnabled(True)
        else:
            self.toolButton_preset_load.setEnabled(False)

    @QtCore.pyqtSlot()
    def on_proceed(self):
        """User is done and clicked the proceed button

        This will first trigger a creation of the draft dataset
        on DCOR. Then, the job is enqueued in the parent
        """
        # We should only proceed if we have resources
        if self.rvmodel.rowCount() == 0:
            QtWidgets.QMessageBox.critical(self, "No resources selected",
                                           "Please add at least one resource.")
            return
        # Checking for duplicate resources is the responsibility of
        # DCOR-Aid, because we are skipping existing resources in
        # dcoraid.upload.job.UploadJob.task_upload_resources.
        if not self.rvmodel.filenames_are_unique():
            QtWidgets.QMessageBox.critical(
                self,
                "Resource names not unique",
                "Please make sure that all resources have a unique file name. "
                + "Resources with identical file names are not supported by "
                + "DCOR.")
            return
        if not self.rvmodel.filenames_were_edited():
            choice = QtWidgets.QMessageBox.question(
                self,
                "Only raw file names",
                "You did not rename any of the resource file names. Was this "
                + "intentional? Using generic file names is discouraged. "
                + "Please select 'No' if you would like to change things. "
                + "Select 'Yes' to proceed without changes - be aware that "
                + "you will not be able to change anything later on.")
            if choice != QtWidgets.QMessageBox.Yes:
                return
        if not self.rvmodel.supplements_were_edited():
            choice = QtWidgets.QMessageBox.question(
                self,
                "No supplementary resource metadata",
                "You did not specify any supplementary resource metadata. "
                + "Would you like to proceed anyway? The editing options "
                + "appear when you click on a resource. Please select 'No' "
                + "if you would like to go back (recommended). Select 'Yes' "
                + "to proceed without changes (no future changes possible).")
            if choice != QtWidgets.QMessageBox.Yes:
                return

        # Try to create the dataset and display any issues with the metadata
        try:
            data = create_dataset(dataset_dict=self.assemble_metadata(),
                                  api=self.api.copy())
        except BaseException:
            msg = QtWidgets.QMessageBox()
            msg.setIcon(QtWidgets.QMessageBox.Critical)
            msg.setText("It was not possible to create a dataset draft. "
                        "If this is not a connection problem, please consider "
                        "creating an <a href='"
                        "https://github.com/DCOR-dev/DCOR-Aid/issues"
                        "'>issue on GitHub</a>.")
            msg.setWindowTitle("Dataset creation failed")
            msg.setDetailedText(tb.format_exc())
            msg.exec_()
            return
        self.setHidden(True)
        # Remember the dataset identifier
        self.dataset_dict = data
        self.dataset_id = data["id"]
        # signal that we are clear to proceed
        self.finished.emit(self)
        self.close()

    @QtCore.pyqtSlot()
    def on_rem_resources(self):
        """Remove the selected resources"""
        sel = self.listView_resources.selectedIndexes()
        self.rvmodel.rem_resources(sel)
        self.listView_resources.clearSelection()

    @QtCore.pyqtSlot()
    def on_selection_changed(self):
        """User changed the ListView selection; Refresh side panel"""
        sel = self.listView_resources.selectedIndexes()
        seltype = self.rvmodel.get_indexes_types(sel)
        # Resource options
        if len(sel) == 1:  # a single resource; show resource options
            self.groupBox_res_info.show()
            # populate
            path, data = self.rvmodel.get_data_for_index(sel[0])
            self.lineEdit_res_filename.blockSignals(True)
            self.lineEdit_res_filename.setText(data["file"]["filename"])
            self.lineEdit_res_filename.blockSignals(False)
            self.lineEdit_res_path.setText(path)
        else:  # hide resource options
            self.groupBox_res_info.hide()
        # Supplement options
        if seltype in ["dc", "mixed"]:
            self.widget_supplement.show()
            # populate
            common = self.rvmodel.get_common_supplements_from_indexes(sel)
            self.widget_schema.set_schema(common)
        else:
            self.widget_supplement.hide()

        # also reset the preset combo box string
        self.comboBox_preset.setCurrentText("")

    @QtCore.pyqtSlot()
    def on_update_resources_model(self):
        """Assemble metadata dictionary and update self.rvmodel"""
        data_dict = {}
        sel = self.listView_resources.selectedIndexes()
        if len(sel) == 0:  # nothing to do
            return
        elif len(sel) == 1:  # update file name
            path = pathlib.Path(self.rvmodel.get_data_for_index(sel[0])[0])
            suffix = path.suffix
            # prevent users from changing the suffix
            fn = self.lineEdit_res_filename.text()
            if not fn.endswith(suffix):
                fn += suffix
                self.lineEdit_res_filename.blockSignals(True)
                self.lineEdit_res_filename.setText(fn)
                self.lineEdit_res_filename.blockSignals(False)
            if fn != path.name:  # only update filename if user changed it
                data_dict["file"] = {"filename": fn}
        # collect supplementary resource data
        schema = self.widget_schema.get_current_schema()
        if schema:  # only update supplement if user made changes
            data_dict["supplement"] = schema
        self.rvmodel.update_resources(sel, data_dict)



from os.path import dirname, realpath, exists
from setuptools import setup, find_packages
import sys


author = u"Paul Müller"
authors = [author]
description = 'GUI for managing data on DCOR'
name = 'dcoraid'
year = "2019"

sys.path.insert(0, realpath(dirname(__file__))+"/"+name)
from _version import version  # noqa: E402

setup(
    name=name,
    author=author,
    author_email='dev@craban.de',
    url='https://github.com/DCOR-dev/DCOR-Aid',
    version=version,
    packages=find_packages(),
    package_dir={name: name},
    include_package_data=True,
    license="GPL v3",
    description=description,
    long_description=open('README.rst').read() if exists('README.rst') else '',
    install_requires=["appdirs",  # cache directory for data compression
                      "dclab[dcor]==0.33.3",  # pin for triage
                      "pyqt5",
                      "requests>=2.13",
                      "requests_toolbelt",  # multipart uploads with progress
                      ],
    python_requires='>=3.6, <4',
    entry_points={"gui_scripts": ['dcoraid = dcoraid.__main__:main']},
    keywords=["RT-DC", "deformability", "cytometry", "zellmechanik"],
    classifiers=['Operating System :: OS Independent',
                 'Programming Language :: Python :: 3',
                 'Intended Audience :: Science/Research',
                 ],
    platforms=['ALL']
    )



import atexit
import pathlib
import shutil
import tempfile

from PyQt5 import QtCore

from ..api import CKANAPI

#: Either a boolean or a path to the server's SSL certificate
_SSL_VERIFY = None


def get_ckan_api():
    """Convenience function for obtaining CKANAPI instance from settings"""
    settings = QtCore.QSettings()
    api_key = settings.value("auth/api key", "")
    server = settings.value("auth/server", "dcor.mpl.mpg.de")
    ssl_verify = setup_certificate_file()
    api = CKANAPI(server=server, api_key=api_key, ssl_verify=ssl_verify)
    return api


def setup_certificate_file():
    global _SSL_VERIFY
    if _SSL_VERIFY is None:
        settings = QtCore.QSettings()
        if settings.value("user scenario") == "medical":
            cert_data = settings.value("auth/certificate")
            tmpdir = tempfile.mkdtemp(prefix="dcoraid_certificate_pinning_")
            cert_path = pathlib.Path(tmpdir) / "server.cert"
            cert_path.write_bytes(cert_data)
            atexit.register(shutil.rmtree, tmpdir, ignore_errors=True)
            _SSL_VERIFY = str(cert_path)
        else:
            _SSL_VERIFY = True
    return _SSL_VERIFY



"""Collect relevant icons to icon theme subdirectories

To install the KDE breeze theme:

    apt install breeze-icon-theme

Fontawesome icons are downloaded from GitHub.

This script must be run on a linux machine. Please make sure
that `local_root` is correct.
"""
import atexit
import collections
import pathlib
import shutil
import tempfile
from xml.dom import minidom

import requests


# The key identifies the theme; each list contains icon names
icons = {
    "breeze": [
        "application-exit",
        "dialog-cancel",
        "dialog-close",
        "dialog-error",
        "dialog-information",
        "dialog-messages",
        "dialog-ok",
        "dialog-ok-apply",
        "dialog-question",
        "dialog-warning",
        "documentinfo",
        "document-open",
        "document-open-folder",
        "document-save",
        "edit-clear",
        "gtk-preferences",
        "messagebox_warning",
        "preferences-activities",
    ],
    "fontawesome": [
        "ban",
        "book",
        "broom",
        "check-double",
        "child",
        "circle",
        "clinic-medical",
        "code-branch",
        "cogs",
        "eye",
        "exclamation-triangle",
        "filter",
        "folder",
        "globe",
        "hat-wizard",
        "hourglass",
        "info",
        "people-arrows",
        "puzzle-piece",
        "redo",
        "slash",
        "street-view",
        "tag",
        "trash",
        "trash-alt",
        "undo",
        "upload",
        "user",
        "user-lock",
        "user-times",
        "users",
    ],
}

# theme index file
index = """[Icon Theme]
Name=DCORAidMix
Comment=Mix of themes for DCOR-Aid

Directories={directories}
"""

# theme file folder item
index_item = """
[{directory}]
Size={res}
Type=Fixed
"""

local_root = pathlib.Path("/usr/share/icons")
web_roots = {
    "fontawesome": collections.OrderedDict(
        solid="https://raw.githubusercontent.com/FortAwesome/"
              + "Font-Awesome/master/svgs/solid/"
    )}


# create a temporary directory
tmpdir = pathlib.Path(tempfile.mkdtemp(prefix="icons_"))
atexit.register(lambda: shutil.rmtree(tmpdir, ignore_errors=True))


def process_svg(svgdata):
    """Set svg color to gray and add a 10% margin"""
    parsed = minidom.parseString(svgdata).childNodes[0]
    # grey color
    parsed.setAttribute("fill", "#232629")
    # larger viewbox
    vb = [int(it) for it in parsed.getAttribute("viewBox").split()]
    margin = vb[2] // 10
    vb = [-margin, -margin, vb[2] + 2*margin, vb[3] + 2*margin]
    parsed.setAttribute("viewBox", " ".join([str(it) for it in vb]))
    return parsed.toxml()


def find_icons(name, theme):
    cands = []
    if theme in web_roots:  # download icons
        for topic in web_roots[theme]:
            filename = name + ".svg"
            url = web_roots[theme][topic]
            response = requests.get(url + filename)
            topic_dir = tmpdir / theme / topic
            topic_dir.mkdir(exist_ok=True, parents=True)
            if response.status_code == 200:
                path = topic_dir / filename
                svgdata = process_svg(response.content.decode())
                with path.open('w') as f:
                    f.write(svgdata)
                relp = path.parent.relative_to(tmpdir)
                cands.append([path, relp])
                break  # we only need one icon
    else:  # use local icons
        svgs = sorted((local_root / theme).rglob("{}.svg".format(name)))
        pngs = sorted((local_root / theme).rglob("{}.png".format(name)))
        for path in svgs + pngs:
            relp = path.parent.relative_to(local_root)
            cands.append([path, relp])
    return cands


if __name__ == "__main__":
    directories = []
    here = pathlib.Path(__file__).parent
    for theme in icons:
        for name in icons[theme]:
            ipaths = find_icons(name, theme)
            if not ipaths:
                print("Could not find {} {}".format(theme, name))
                continue
            for ipath, relp in ipaths:
                dest = here / relp
                directories.append(str(relp))
                dest.mkdir(exist_ok=True, parents=True)
                shutil.copy(ipath, dest)

    with (here / "index.theme").open("w") as fd:
        directories = sorted(set(directories))
        fd.write(index.format(directories=",".join(
            ["dcoraid"] + directories)))
        # Shape-Out icons
        fd.write(index_item.format(directory="dcoraid", res="16"))
        # theme icons
        for dd in directories:
            for res in ["16", "22", "24", "32", "64", "128"]:
                if res in str(dd):
                    break
            else:
                res = "64"
            fd.write(index_item.format(directory=dd, res=res))



from collections import OrderedDict
import pkg_resources

from PyQt5 import QtCore, QtWidgets, uic


class FilterBase(QtWidgets.QWidget):
    #: The user selection has changed
    selection_changed = QtCore.pyqtSignal(list)

    def __init__(self, *args, **kwargs):
        """Filter view widget with title, edit, checkbox, and table
        """
        super(FilterBase, self).__init__(*args, **kwargs)
        path_ui = pkg_resources.resource_filename("dcoraid.gui.dbview",
                                                  "filter_base.ui")
        uic.loadUi(path_ui, self)
        # trigger user selection change signal
        self.listWidget.itemSelectionChanged.connect(self.on_item_selected)

    def get_item_keys(self, selected=False):
        """Return the keys of the dictionary items"""
        if selected:
            keys = []
            for ii, key in enumerate(self.item_dict):
                if self.listWidget.item(ii).isSelected():
                    keys.append(key)
        else:
            keys = list(self.item_dict.keys())
        return keys

    @QtCore.pyqtSlot()
    def on_item_selected(self):
        keys = self.get_item_keys(selected=True)
        self.selection_changed.emit(keys)

    def select_item_keys(self, keys):
        """Select entries in `keys`, all other entries are deselected"""
        for ii, key in enumerate(self.item_dict):
            if key in keys:
                self.listWidget.item(ii).setSelected(True)
            else:
                self.listWidget.item(ii).setSelected(False)

    def set_items(self, item_dict):
        """Set the current list items

        Parameters
        ----------
        item_dict: dict
            A dictionary {"key": "displayed text", ...} of the
            items to be displayed.
        """
        if not isinstance(item_dict, dict):
            raise ValueError(
                "`item_dict` must be dict, got '{}'!".fromat(item_dict))
        self.item_dict = OrderedDict(item_dict)
        self.listWidget.clear()
        for key in self.item_dict:
            self.listWidget.addItem(self.item_dict[key])



from PyQt5 import QtCore, QtWidgets

from .widget_supplement_item import RSSItem, RSSTagsItem, TitleItem


#: These items displayed differently to the user and may populate other keys
MAGIC_ITEMS = {
    "identifiers": {
        "tags": {
            "widget": RSSTagsItem,  # use this instead of RSSItem
            "setters": [
                {"section": "identifiers",
                 "key": "compounds",
                 "func": "get_compounds"
                 },
                {"section": "identifiers",
                 "key": "labels",
                 "func": "get_labels"
                 },
                {"section": "identifiers",
                 "key": "tags",
                 "func": "get_tags"
                 },
            ]
        }
    }
}

#: These items are defined by magical items (or are just ignored)
HIDDEN_ITEMS = {
    "identifiers": ["antibodies", "fluorophores"]
}


class SchemaWidget(QtWidgets.QWidget):
    schema_changed = QtCore.pyqtSignal()

    def __init__(self, *args, **kwargs):
        """Widget containing groups of """
        super(SchemaWidget, self).__init__(*args, **kwargs)
        self.verticalLayout = QtWidgets.QVBoxLayout(self)
        self.schema_widgets = {}

    def get_current_schema(self):
        """Return the current schema dict for storage in the model

        The schema also contains magic items (see :const:`MAGIC_ITEMS`)
        which are super-items that contain information about more than
        one item.
        """
        schema = {}
        for sec in self.schema_widgets:
            for widget in self.schema_widgets[sec]:
                if widget.checkBox.isChecked():
                    if sec not in schema:
                        schema[sec] = {}
                    key = widget.rss_dict["key"]
                    # We extract all relevant information from magic items
                    if sec in MAGIC_ITEMS and key in MAGIC_ITEMS[sec]:
                        schema[sec]["MAGIC_{}".format(
                            key)] = widget.get_value()
                        for sr in MAGIC_ITEMS[sec][key]["setters"]:
                            fc = getattr(widget, sr["func"])
                            schema[sr["section"]][sr["key"]] = fc()
                    else:
                        schema[sec][key] = widget.get_value()
        return schema

    @QtCore.pyqtSlot(str, str, object, bool)
    def on_hide_show_items(self, section, key, value, enabled):
        """Send a value_changed signal to all widgets

        ...so they can decide whether they are visible or not.
        """
        for sec in self.schema_widgets:
            for item in self.schema_widgets[sec]:
                item.on_schema_key_toggled(section, key, value, enabled)
        for label in self.schema_title_widgets:
            label.on_schema_key_toggled(section, key, value, enabled)

    def populate_schema(self, schema_dict):
        """Create all widgets corresponding to the schema

        Omits any :const:`HIDDEN_ITEMS` and creates special widgets
        for :const:`MAGIC_ITEMS`.
        """
        self.schema_title_widgets = []
        for sec in schema_dict:
            widget_list = []
            label = TitleItem(schema_dict[sec].get("requires", {}),
                              schema_dict[sec]["name"])
            self.schema_title_widgets.append(label)
            self.verticalLayout.addWidget(label)
            for item in schema_dict[sec]["items"]:
                key = item["key"]
                # Update item requirements with section requirements
                if "requires" not in item:
                    item["requires"] = schema_dict[sec].get("requires", {})
                if sec in MAGIC_ITEMS and key in MAGIC_ITEMS[sec]:
                    # is a special widget (e.g. tags)
                    wrss = MAGIC_ITEMS[sec][key]["widget"](item,
                                                           section=sec,
                                                           parent=self)
                elif sec in HIDDEN_ITEMS and key in HIDDEN_ITEMS[sec]:
                    # should not be displayed
                    continue
                else:
                    wrss = RSSItem(item, section=sec, parent=self)
                self.verticalLayout.addWidget(wrss)
                widget_list.append(wrss)
                wrss.value_changed.connect(self.schema_changed)
                wrss.value_changed.connect(self.on_hide_show_items)
            self.schema_widgets[sec] = widget_list

    def set_schema(self, schema_dict):
        """Apply schema to all widgets

        If there are any special :const:`MAGIC_ITEMS` widgets,
        then the `schema_dict` should contain the corresponding
        `MAGIC_key` key (which is used to set the value of the
        special widget).
        """
        self.blockSignals(True)
        for sec in self.schema_widgets:
            for wrss in self.schema_widgets[sec]:
                key = wrss.rss_dict["key"]
                match = sec in schema_dict and key in schema_dict[sec]
                wrss.check(match)
                if match:
                    # Note that hidden items don't show up here since
                    # we iterate over the widgets.
                    if sec in MAGIC_ITEMS and key in MAGIC_ITEMS[sec]:
                        value = schema_dict[sec]["MAGIC_{}".format(key)]
                    else:
                        value = schema_dict[sec][key]
                    wrss.set_value(value)
        self.blockSignals(False)
        self.update_visible_widgets()

    def update_visible_widgets(self):
        """Cause all schema data widgets to emit their values

        This effectively updates the visibility of all schema
        widgets (including titles).
        """
        # Force all updates
        for sec in self.schema_widgets:
            for widget in self.schema_widgets[sec]:
                widget.emit_value()



# flake8: noqa: F401
from .dataset import activate_dataset, add_resource, create_dataset, \
    remove_draft, remove_all_drafts, resource_exists, resource_sha256_sums
from .queue import UploadQueue, UploadJob
